import datetime
import string
from etk.timeseries.annotation.utility import date_cell, number_cell, text_cell, empty_cell, blank_cell, special_keywords
import logging
import re

class simple_tag_classifier:

    def get_date_tag(self):
        return date_cell

    def tag_sheet(self, raw_sheet):
        tags = []
        i = 0
        for row in raw_sheet.rows():
            tags.append([])
            for col in row:
                tags[i].append(self.find_class(col))
            i += 1
        return tags

    def find_class(self, cell):
            if type(cell) == datetime.date:
                return {date_cell}
            if type(cell) == datetime.datetime:
                return {date_cell}
            if type(cell) == float:
                return {number_cell}
            if type(cell) == int or type(cell) == long:
                if cell < 2020 and cell > 1900:
                    return {date_cell, number_cell}
                else:
                    return {number_cell, text_cell}
            if type(cell) == str or type(cell) == unicode:
                if cell == '-':
                    return [blank_cell]
                if len(string.replace(cell, ' ', '')) == 0:
                    return {empty_cell}
                if self.possible_date_presentation(cell):
                    return {date_cell, text_cell}
                float_flag = False
                try:
                    a = float(cell)
                    float_flag = True
                    a = int(cell)
                    return {number_cell, text_cell}
                except:
                    if float_flag:
                        return {number_cell}
                return {text_cell}
            logging.error("wow this cell has no tag!!!!!! " + str(cell) + " " + type(cell))
            
    def possible_date_presentation(self, cell):
        try:
            x = int(cell)
            if x > 1900 and x <= 2020:  # I should add the current date parameter
                return True
        except:
            for word in special_keywords:
                if cell.lower().strip() == word:
                    return True

            cell_modified = string.replace(cell, '-', '/')
            cell_modified = string.replace(cell_modified, '.', '/')
            cell_modified = string.replace(cell_modified, '/', ' ')

            #    print splitted_cell
            if len(string.replace(cell_modified, ' ', '')) == 0:
                return False

            splitted_cell = (cell_modified.strip()).split()

            if len(splitted_cell) > 3:
                return False
            if len(splitted_cell) == 1:

                matchObj = re.match(r'(\d{4})[^\d].*', cell, re.M | re.I)
                if matchObj:
                    # print cell + 'matched'
                    if int(matchObj.group(1)) > 1900 and int(matchObj.group(1))<2020:
                        return True
                matchObj = re.match(r'.*[^\d](\d{4})', cell, re.M | re.I)
                if matchObj:
                    if int(matchObj.group(1)) > 1900 and int(matchObj.group(1))<2020:
                        return True
                # print 'not date: ' + cell
                return False

            for part in splitted_cell:
                if part.lower() in special_keywords:
                    continue
                try:
                    x = int(part)
                    if x == 0:
                        return False
                    if x >= 1900 and x < 2020:
                        continue
                    if x <= 31:
                        continue
                    else:
                        return False
                except:
                    return False
            return True


import unittest
import json
from etk.knowledge_graph.schema import KGSchema
from etk.etk import ETK
from etk.etk_exceptions import KGValueError
from datetime import date, datetime
from etk.knowledge_graph.ontology import Ontology
from etk.knowledge_graph.namespacemanager import DIG
from etk.knowledge_graph.node import URI, BNode, Literal, LiteralType
from etk.knowledge_graph.subject import Subject
import rdflib


class TestKnowledgeGraph(unittest.TestCase):
    def setUp(self):
        sample_doc = {
            "projects": [
                {
                    "name": "etk",
                    "description": "version 2 of etk, implemented by Runqi12 Shao, Dongyu Li, Sylvia lin, Amandeep and others.",
                    "members": ["dongyu", "amandeep", "sylvia", "Runqi12"],
                    "date": "2007-12-05",
                    "place": "columbus:georgia:united states:-84.98771:32.46098",
                    "s": "segment_test_1"
                },
                {
                    "name": "rltk",
                    "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students.",
                    "members": ["mayank", "yixiang"],
                    "date": ["2007-12-05T23:19:00"],
                    "cost": -3213.32,
                    "s": "segment_test_2"
                }
            ]
        }
        kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

        etk = ETK(kg_schema)
        self.doc = etk.create_document(sample_doc, doc_id="http://isi.edu/default-ns/projects")

    def test_add_segment_kg(self) -> None:
        sample_doc = self.doc
        segments = sample_doc.select_segments("projects[*].s")
        for segment in segments:
            sample_doc.kg.add_value("segment", segment.value)
        expected_segments = ["segment_test_1", "segment_test_2"]
        self.assertIn(sample_doc.kg.value["segment"][0]["key"], expected_segments)
        self.assertIn(sample_doc.kg.value["segment"][1]["key"], expected_segments)
        # self.assertIn('provenances', sample_doc.value)
        # provenances = sample_doc.value['provenances']
        # self.assertEqual(len(provenances), 2)
        # self.assertEqual(provenances[0]['reference_type'], 'location')

    def test_KnowledgeGraph(self) -> None:
        sample_doc = self.doc

        try:
            for member in sample_doc.select_segments("projects[*].members[*]"):
                sample_doc.kg.add_value("developer", member.value)
        except KGValueError:
            pass

        # TODO: for property with range: xsd:date, auto convert to literal(type=date)
        # try:
        #     for date_ in sample_doc.select_segments("projects[*].date[*]"):
        #         sample_doc.kg.add_value("test_date", date_.value)
        # except KGValueError:
        #     pass

        # TODO: for date obj, auto convert to literal(type=date)
        # try:
        #     sample_doc.kg.add_value("test_add_value_date", date(2018, 3, 28))
        #     sample_doc.kg.add_value("test_add_value_date", {})
        #     sample_doc.kg.add_value("test_add_value_date", datetime(2018, 3, 28, 1, 1, 1))
        # except KGValueError:
        #     pass

        try:
            for place in sample_doc.select_segments("projects[*].place"):
                sample_doc.kg.add_value("test_location", place.value)
        except KGValueError:
            pass

        expected_developers = [
            {
                "key": "amandeep",
                "value": "amandeep"
            },
            {
                "key": "dongyu",
                "value": "dongyu"
            },
            {
                "key": "mayank",
                "value": "mayank"
            },
            {
                "key": "runqi12",
                "value": "Runqi12"
            },
            {
                "key": "sylvia",
                "value": "sylvia"
            },
            {
                "key": "yixiang",
                "value": "yixiang"
            }
        ]

        expected_date = [
            {
                "value": "2007-12-05T00:00:00",
                "key": "2007-12-05T00:00:00"
            },
            {
                "value": "2007-12-05T23:19:00",
                "key": "2007-12-05T23:19:00"
            }
        ]

        expected_add_value_date = [
            {
                "value": "2018-03-28",
                "key": "2018-03-28"
            },
            {
                "value": "2018-03-28T01:01:01",
                "key": "2018-03-28T01:01:01"
            }
        ]

        expected_location = [
            {
                "value": "columbus:georgia:united states:-84.98771:32.46098",
                "key": "columbus:georgia:united states:-84.98771:32.46098"
            }
        ]

        self.assertEqual(expected_developers,
                         sorted(sample_doc.kg.value["developer"], key=lambda x: x['key']))
        # self.assertEqual(expected_date, sample_doc.kg.value["test_date"])
        self.assertEqual(expected_location, sample_doc.kg.value["test_location"])
        # self.assertEqual(expected_add_value_date, sample_doc.kg.value["test_add_value_date"])

    def test_add_value_empty(self):
        pass
        # TODO: auto convert 0.0 to Literal(type=decimal) and convert back with kg.value
        # self.doc.kg.add_value('test_zero', 0.0)
        # self.assertEqual(self.doc.kg.value['test_zero'][0]['value'], 0.0)


class TestKnowledgeGraphWithOntology(unittest.TestCase):
    def setUp(self):
        ontology_content = '''
            @prefix : <http://dig.isi.edu/ontologies/dig/> .
            @prefix dig: <http://dig.isi.edu/ontologies/dig/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix schema: <http://schema.org/> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :Person a owl:Class ;
                rdfs:subClassOf :Actor, :Biological_Object ;
                :common_properties :label, :title, :religion ; .
            :has_name a owl:DatatypeProperty ;
                schema:domainIncludes :Person ;
                schema:rangeIncludes xsd:string ; .
            :has_child a owl:ObjectProperty ;
                schema:domainIncludes :Person ;
                schema:rangeIncludes :Person ; .
            '''
        kg_schema = KGSchema()
        kg_schema.add_schema(ontology_content, 'ttl')
        etk = ETK(kg_schema=kg_schema)
        self.doc = etk.create_document(dict(), doc_id='http://xxx/1', type_=[URI('dig:Person')])

    def test_valid_kg(self):
        kg = self.doc.kg
        self.assertIsInstance(kg._g, rdflib.Graph)
        self.assertEqual(len(kg._g), 1)
        self.assertIn((rdflib.URIRef('http://xxx/1'), rdflib.RDF.type, DIG.Person), kg._g)

    def test_valid_ontology(self):
        onto = self.doc.etk.kg_schema.ontology
        self.assertIsInstance(onto._g, rdflib.Graph)
        self.assertEqual(len(onto._g), 12)

    def test_add_types(self):
        kg = self.doc.kg
        self.assertIn((rdflib.URIRef('http://xxx/1'), rdflib.RDF.type, DIG.Person), kg._g)
        kg.add_types(URI('dig:Male'))
        self.assertIn((rdflib.URIRef('http://xxx/1'), rdflib.RDF.type, DIG.Male), kg._g)
        kg.add_types([URI('dig:Human')])
        self.assertIn((rdflib.URIRef('http://xxx/1'), rdflib.RDF.type, DIG.Human), kg._g)

    def test_add_subject(self):
        kg = self.doc.kg
        subject = Subject(URI('http://xxx/1'))
        subject.add_property(URI('dig:name'), Literal('Jack', lang='en'))
        subject.add_property(URI('dig:age'), Literal('18', type_=LiteralType.int))
        friend = Subject(BNode())
        friend.add_property(URI('rdf:type'), Literal('dig:Person'))
        friend.add_property(URI('dig:name'), Literal('José', lang='es'))
        friend.add_property(URI('dig:age'), Literal('19', type_=LiteralType.int))
        friend.add_property(URI('dig:friend'), subject)
        subject.add_property(URI('dig:friend'), friend)
        kg.add_subject(subject)
        self.assertEqual(len(kg._g), 8)
        self.assertIn((rdflib.URIRef('http://xxx/1'), DIG.name, rdflib.Literal('Jack', lang='en')), kg._g)



# -*- coding: utf-8 -*-
from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction

from bs4 import BeautifulSoup
from bs4.element import Comment
import re
import copy
import pandas as pd
from zipfile import ZipFile
from requests import get

class Toolkit:
    @staticmethod
    def create_table_array(t, put_extractions=False):
        rows = t['rows']
        tt = []
        max_cols = t['features']['max_cols_in_a_row']
        for r in rows:
            new_r = ['' for xx in range(max_cols)]
            for i, c in enumerate(r['cells']):
                text = c['text']
                text = text.lower()
                text = text.strip()
                if put_extractions and 'data_extraction' in c:
                    data_extractions = c['data_extraction']
                    for key in data_extractions.keys():
                        text += ' DUMMY' + key.upper()
                new_r[i] = text.strip()
            tt.append(new_r)
        return tt

    @staticmethod
    def regulize_cells(t):
        for r in t:
            for i in range(len(r)):
                r[i] = re.sub('[0-9]', 'NUM', r[i])
                # for x in re.findall('([0-9])', r[i]):
                #     int_x = int(x)
                #     if int_x < 5:
                #         r[i] = re.sub(x, 'SSSS', r[i])
                #     else:
                #         r[i] = re.sub(x, 'LLLL', r[i])
                for x in re.findall('([a-z][a-z][a-z]+@)', r[i]):
                    r[i] = re.sub(x, 'EMAILNAME ', r[i])

    @staticmethod
    def clean_cells(t): # modifies t
        for r in t:
            for i in range(len(r)):
                r[i] = re.sub(r'[^\x00-\x7F]',' ', r[i]) #remove unicodes
                # r[i] = re.sub('[\'"]', '', r[i]) #remove annoying puncts
                r[i] = re.sub('[^\s\w\.\-\$_%\^&*#~+@"\']', ' ', r[i]) #remove annoying puncts
                for x in re.findall('(\.[a-z])', r[i]):
                    r[i] = re.sub('\.{0}'.format(x[1]), ' {0}'.format(x[1]), r[i])
                r[i] = re.sub('\s+', ' ', r[i])
                r[i] = r[i].strip()


class EntityTableDataExtraction(Extractor):
    def __init__(self) -> None:
        Extractor.__init__(self,
                           input_type=InputType.OBJECT,
                           category="data",
                           name="DigEntityTableDataExtractor")
        self.glossaries = dict()

    def add_glossary(self, glossary: List[str], attr_name: str) -> None:
        """
        Adds a glossary for the given attribute name
        :param glossary: a list of possible mentions of the attribute name
        :param attr_name: the attribute name (field name)
        """
        self.glossaries[attr_name] = glossary

    def wrap_value_with_context(self, value: dict, field_name: str, start: int=0, end: int=0) -> Extraction:
        """Wraps the final result"""
        return Extraction(value, self.name, start_token=start, end_token=end, tag=field_name)

    def extract(self, table: dict) -> List[Extraction]:
        """
        :param table: a table extracted by table extractor, as a json object
        :return: list of all extractions from the input table
        """
        if table['features']['max_cols_in_a_row'] != 2 and table['features']['no_of_rows'] < 2:
            return []
        results = list()
        for row in table['rows']:
            if len(row['cells']) != 2:
                continue
            text = [row['cells'][0]['text'], row['cells'][1]['text']]
            for field_name in self.glossaries.keys():
                if self.cell_matches_dict(text[0], self.glossaries[field_name]):
                    results.append(self.wrap_value_with_context(text[1], field_name))
                if self.cell_matches_dict(text[1], self.glossaries[field_name]):
                    results.append(self.wrap_value_with_context(text[0], field_name))
        return results

    def cell_matches_dict(self, cell_text: str, glossary: List[str]) -> bool:
        if any([self.cell_matches_text(cell_text, x) for x in glossary]):
            return True
        return False

    def cell_matches_text(self, cell_text: str, text: str) -> bool:
        cell_text = cell_text.lower()
        text = text.lower()
        if text in cell_text and float(len(cell_text))/float(len(text)) < 1.5:
            return True
        return False


class TableExtraction:
    @staticmethod
    def is_data_cell(cell):
        if cell.table:
            return False
        return True

    @staticmethod
    def is_data_row(row):
        if row.table:
            return False
        cell = row.findAll('th', recursive=False)
        cell.extend(row.findAll('td', recursive=False))
        for td in cell:
            if TableExtraction.is_data_cell(td) == False:
                return False
        return True

    @staticmethod
    def get_data_rows(table):
        data_rows = []
        rows = table.findAll('tr', recursive=False)
        if table.thead:
            rows.extend(table.thead.findAll('tr', recursive=False))
        if table.tbody:
            rows.extend(table.tbody.findAll('tr', recursive=False))
        for tr in rows:
            if TableExtraction.is_data_row(tr):
                data_rows.append(tr)
        return data_rows

    @staticmethod
    def is_data_table(table, k):
        rows = TableExtraction.get_data_rows(table)
        if len(rows) > k:
            return rows
        else:
            return False

    @staticmethod
    def mean(numbers):
        """ Computes mean of a list of numbers """
        return float(sum(numbers)) / max(len(numbers), 1)

    @staticmethod
    def _ss(data):
        """Return sum of square deviations of sequence data."""
        c = TableExtraction.mean(data)
        ss = sum((x-c)**2 for x in data)
        return ss

    @staticmethod
    def pstdev(data):
        """Calculates the population standard deviation."""
        n = len(data)
        if n < 2:
            return 0
            # raise ValueError('variance requires at least two data points')
        ss = TableExtraction._ss(data)
        pvar = ss/n # the population variance
        return pvar**0.5

    @staticmethod
    # Check if a string contains a digit
    def contains_digits(d):
        _digits = re.compile('\d')
        return bool(_digits.search(d))

    @staticmethod
    def gen_context(seq):
        seen = set()
        seen_add = seen.add
        uniq_list = [x for x in seq if not (x in seen or seen_add(x))]
        if len(uniq_list) > 5:
            uniq_list = uniq_list[:5]
        uniq_list = [x.replace("\t", "").replace("\r", "").replace("\n", "").strip() for x in uniq_list]
        if '' in uniq_list:
            uniq_list.remove('')
        if ' ' in uniq_list:
            uniq_list.remove(' ')
        return uniq_list

    @staticmethod
    def convert_to_csv(jobj, outpath):
        with ZipFile(outpath, 'w') as myzip:
            for ti, t in enumerate(jobj['tables']):
                with myzip.open(f't{ti}.csv', 'w') as myfile:
                    tarr = [[c['text'] if c is not None else '' for c in r['cells']] for r in t['rows']]
                    csv = pd.DataFrame(tarr).to_csv(index=False, header=False)
                    myfile.write(csv.encode('utf-8'))
                myfile.close()

    @staticmethod
    def extract(input_data, expand_merged_cells=False, 
                input_format='html', output_format='json', 
                min_data_rows=1, output_path=None):
        """extract tables from input data.

        Args:
            input_data (str): input url/html.
            expand_merged_cells (bool): replicate the value of merged cells.
            input_format (str): the form of input data, {html, pdf}
            output_format (str): output format, {json, csv}
            min_data_rows (int): minimum number of rows for data tables.
            output_path (str): path to output file if output format is csv.

        Returns:
            bool: json object if output_format is json, None otherwise.
            
        Examples:

        >>> url = 'https://en.wikipedia.org/wiki/United_States'
        >>> html = requests.get(url).text
        >>> te = TableExtraction()
        >>> res = te.extract(html, expand_merged_cells=False, 
                        output_format='csv', output_path='test.zip')

        >>> res = te.extract(html)['tables'][0]['text']
        United States of America | \nFlag Coat of arms | \nMot ....
        
        >>> url = 'https://www.imf.org/~/media/Files/Publications/WP/wp1715.ashx'
        >>> te.extract(url, input_format='pdf',
                        expand_merged_cells=False, output_format='csv', 
                        output_path='test.csv')
        """
        if input_format == 'html':
            json_output = TableExtraction.extract_html(input_data, expand_merged_cells, min_data_rows)
            if output_format == 'json':
                return json_output
            elif output_format == 'csv':
                return TableExtraction.convert_to_csv(json_output, output_path)
        elif input_format == 'pdf':
            import tabula
            res = get(input_data)
            with open(output_path+".pdf", 'wb') as tempfile:
                tempfile.write(res.content)
            tempfile.close()
            tabula.convert_into(output_path+".pdf", output_path, output_format="zip", pages='all')

    @staticmethod
    def extract_html(html_doc, expand_merged_cells, min_data_rows=1):
        soup = BeautifulSoup(html_doc, 'html5lib')
        result_tables = list()
        tables = soup.findAll('table')
        for table_i, table in enumerate(tables):
            tdcount = 0
            max_tdcount = 0
            img_count = 0
            href_count = 0
            inp_count = 0
            sel_count = 0
            colspan_count = 0
            colon_count = 0
            len_row = 0
            table_data = ""
            data_table = dict()
            row_list = list()
            rows = TableExtraction.is_data_table(table, min_data_rows)
            if rows != False:
                features = dict()
                row_len_list = list()
                avg_cell_len = 0
                avg_row_len_dev = 0
                
                num_rows = 0
                num_cols = 0
                max_cols = 0
                row_shift = 0
                
                ## detect merged cells and find table demension
                merged_cells = []
                row_spans = dict()
                for ri, row in enumerate(rows):
                    rshift = 0
                    num_rows += 1
                    num_cols = 0
                    col_shift = 0
                    ci = 0
                    for ci, c in enumerate(row.findAll(['td', 'th'])):
                        num_cols += 1
                        ci += col_shift+rshift
                        # shift the col index if there are any spanning cells above it
                        while ci in row_spans:
                            if row_spans[ci] <= 0:
                                del row_spans[ci]
                            else:
                                rshift += 1
                                row_spans[ci] -= 1
                                ci += 1

                        cspan = c.get('colspan')
                        rspan = c.get('rowspan')

                        # record spanned cell for later use
                        if cspan is not None and rspan is not None:
                            cspan = int(cspan)
                            rspan = int(rspan)
                            col_shift += cspan-1
                            for ii in range(ci, ci+cspan):
                                row_spans[ii] = rspan-1
                            merged_cells.append((ri, ri+rspan, ci, ci+cspan))
                        elif cspan is not None:
                            cspan = int(cspan)
                            col_shift += cspan-1
                            merged_cells.append((ri, ri+1, ci, ci+cspan))
                        elif rspan is not None:

                            rspan = int(rspan)
                            row_spans[ci] = rspan-1
                            merged_cells.append((ri, ri+rspan, ci, ci+1))
                    if max_cols < num_cols:
                        max_cols = num_cols
                    # update rowspan dict for columns not seen in this iteration
                    for k, v in row_spans.items():
                        if k > ci:
                            row_spans[k] -= 1
                if len(merged_cells) > 0:
                    max_cols = max(max_cols, max([x[3] for x in merged_cells]))
                    num_rows = max(len(rows), max([x[1] for x in merged_cells]))
                else:
                    num_rows = len(rows)
                ## create table array
                row_spans = dict()
                for ri, row in enumerate(rows):
                    row_data = ' '.join(row.stripped_strings)
                    row_data = row_data.replace("\\t", "").replace("\\r", "").replace("\\n", "")
                    
                    row_len_list.append(len(row_data))
                    row_tdcount = len(row.findAll('td')) + len(row.findAll('th'))
                    if row_tdcount > max_tdcount:
                        max_tdcount = row_tdcount
                    tdcount += row_tdcount
                    img_count += len(row.findAll('img'))
                    href_count += len(row.findAll('a'))
                    inp_count += len(row.findAll('input'))
                    sel_count += len(row.findAll('select'))
                    colspan_count += row_data.count("colspan")
                    colon_count += row_data.count(":")
                    len_row += 1
                    table_data += str(row)
                    
                    row_dict = dict()
                    newr = [None]*max_cols
                    shift = 0
                    rshift = 0
                    ci = 0
                    
                    for i, c in enumerate(row.findAll(['td', 'th'])):
                        ci = i+shift+rshift
                        while ci in row_spans:
                            if row_spans[ci] <= 0:
                                del row_spans[ci]
                            else:
                                rshift += 1
                                row_spans[ci] -= 1
                                ci += 1
                        for br in c.find_all("br"):
                            br.replace_with(" ")
                        for br in c.find_all("script"):
                            br.decompose()
                        cell_dict = dict()
                        cell_dict["cell"] = str(c)
                        cell_dict["text"] = ' '.join(c.stripped_strings)
                        cell_dict["id"] = 'row_{0}_col_{1}'.format(ri, ci)
                        
                        avg_cell_len += len(cell_dict["text"])
                        newr[ci] = cell_dict
                        cspan = c.get('colspan')
                        rspan = c.get('rowspan')

                        if cspan is not None and rspan is not None:
                            cspan = int(cspan)
                            rspan = int(rspan)
                            shift += cspan-1
                            for ii in range(ci, ci+cspan):
                                row_spans[ii] = rspan-1
                        elif rspan is not None:
                            rspan = int(rspan)
                            row_spans[ci] = rspan-1
                        elif cspan is not None:
                            cspan = int(cspan)
                            shift += cspan-1
                    for i in range(ci+1, max_cols):
                        if i in row_spans:
                            row_spans[i] -= 1
                    
                    avg_row_len_dev += TableExtraction.pstdev([len(x["text"]) if x else 0 for x in newr])
                    row_dict["cells"] = newr
                    row_dict["text"] = TableExtraction.row_to_text(newr)
                    row_dict["html"] = TableExtraction.row_to_html(newr)
                    row_dict["id"] = "row_{}".format(ri)
                    row_list.append(row_dict)
                if expand_merged_cells:
                    # replicate merged cells
                    N = len(row_list)
                    M = max_cols
                    for m in merged_cells:
                        if row_list[m[0]]['cells'][m[2]] is None:
                            print(m)
                        for ii in range(m[0], min(m[1], N)):
                            for jj in range(m[2], min(m[3], M)):
                                if ii == m[0] and jj == m[2]:
                                    continue
                                row_list[ii]['cells'][jj] = copy.deepcopy(row_list[m[0]]['cells'][m[2]])
                                row_list[ii]['cells'][jj]['id'] += '_span_row{}_col{}'.format(ii,jj)
                
                # To avoid division by zero
                if len_row == 0:
                    tdcount = 1

                features['merged_cells'] = merged_cells
                features["no_of_rows"] = len_row
                features["no_of_cells"] = tdcount
                features["max_cols_in_a_row"] = max_tdcount
                features["ratio_of_img_tags_to_cells"] = img_count*1.0/tdcount
                features["ratio_of_href_tags_to_cells"] = href_count*1.0/tdcount
                features["ratio_of_input_tags_to_cells"] = inp_count*1.0/tdcount
                features["ratio_of_select_tags_to_cells"] = sel_count*1.0/tdcount
                features["ratio_of_colspan_tags_to_cells"] = colspan_count*1.0/tdcount
                features["ratio_of_colons_to_cells"] = colon_count*1.0/tdcount
                features["avg_cell_len"] = avg_cell_len*1.0/tdcount
                features["avg_row_len"] = TableExtraction.mean(row_len_list)
                features["avg_row_len_dev"] = avg_row_len_dev*1.0/max(len_row, 1)

                avg_col_len = 0
                avg_col_len_dev = 0
                no_of_cols_containing_num = 0
                no_of_cols_empty = 0

                if colspan_count == 0.0 and \
                    len_row != 0 and \
                    (tdcount/(len_row * 1.0)) == max_tdcount:
                    col_data = dict()
                    for i in range(max_tdcount):
                        col_data['c_{0}'.format(i)] = []
                    soup_col = BeautifulSoup(table_data, 'html.parser')
                    for row in soup_col.findAll('tr'):
                        h_index = 0
                        h_bool = True
                        for col in row.findAll('th'):
                            col_content = ' '.join(col.stripped_strings)
                            h_bool = False
                            if col_content is None:
                                continue
                            else:
                                col_data['c_{0}'.format(h_index)].append(col_content)
                            h_index += 1
                        d_index = 0
                        if(h_index == 1 and h_bool == False):
                            d_index = 1
                        for col in row.findAll('td'):
                            col_content = ' '.join(col.stripped_strings)
                            if col_content is None:
                                d_index += 1
                                continue
                            else:
                                col_data['c_{0}'.format(d_index)].append(col_content)
                            d_index += 1

                    for key, value in col_data.items():
                        whole_col = ' '.join(value)
                        # avg_cell_len += float("%.2f" % mean([len(x) for x in value]))
                        avg_col_len += sum([len(x) for x in value])
                        avg_col_len_dev += TableExtraction.pstdev([len(x) for x in value])
                        no_of_cols_containing_num += 1 if TableExtraction.contains_digits(whole_col) is True else 0
                        # features["column_" + str(key) + "_is_only_num"] = whole_col.isdigit()
                        no_of_cols_empty += 1 if (whole_col == '') is True else 0
                # To avoid division by zero
                if max_tdcount == 0:
                    max_tdcount = 1
                features["avg_col_len"] = avg_col_len*1.0/max_tdcount
                features["avg_col_len_dev"] = avg_col_len_dev/max_tdcount
                features["no_of_cols_containing_num"] = no_of_cols_containing_num
                features["no_of_cols_empty"] = no_of_cols_empty
                data_table["features"] = features
                data_table["rows"] = row_list
                context_before = ' '.join(TableExtraction.gen_context(table.find_all_previous(string=True)))
                context_after = ' '.join(TableExtraction.gen_context(table.find_all_next(string=True)))
                table_rep = TableExtraction.gen_html(row_list)
                fingerprint = TableExtraction.create_fingerprint(table_rep)
                data_table["context_before"] = context_before
                data_table["context_after"] = context_after
                data_table["fingerprint"] = fingerprint
                data_table['html'] = str(table)
                data_table['text'] = TableExtraction.table_to_text(row_list)
                result_tables.append(data_table)
                table.decompose()
        return dict(tables=result_tables, html_text=TableExtraction.text_from_html(soup))

    @staticmethod
    def create_fingerprint(table):
        table = str(table)
        all_tokens = list(set(re.split('[^\w]+',table)))
        all_tokens = sorted(all_tokens)
        fingerprint = '-'.join(all_tokens)
        return fingerprint

    @staticmethod
    def row_to_html(cells):
        res = '<html><body><table>'
        for i, c in enumerate(cells):
            t = c['cell'] if c else ''
            res += t + '\n'
        res += '</table></body></html>'
        return res

    @staticmethod
    def row_to_text(cells):
        res = ''
        for i, c in enumerate(cells):
            t = c['text'] if c else ''
            res += t
            if i < len(cells)-1:
                res += ' | '
        return res

    @staticmethod
    def table_to_text(rows):
        res = ''
        for row in rows:
            for i, c in enumerate(row['cells']):
                t = c['text'] if c else ''
                res += t
                if i < len(row['cells']) - 1:
                    res += ' | '
            res += '\n'
        return res


    @staticmethod
    def gen_html(row_list):
        """ Return html table string from a list of data rows """
        table = "<table>"
        for row in row_list:
            table += "<tr>"
            cells = row["cells"]
            for c in cells:
                t = c['cell'] if c else ''
                table += t
            table += "</tr>"
        table += "</table>"
        return table

    @staticmethod
    def remove_tables(html_doc, min_data_rows = 1):
        soup = BeautifulSoup(html_doc, 'html.parser')
        tables = soup.findAll('table')
        for table in tables:
            rows = TableExtraction.is_data_table(table, min_data_rows)
            if rows != False:
                table.decompose()

        return soup

    @staticmethod
    def tag_visible(element):
        if element.parent.name in ['style', 'script', 'head', 'title', 'meta', '[document]']:
            return False
        if isinstance(element, Comment):
            return False
        return True

    @staticmethod
    def text_from_html(soup):
        texts = soup.findAll(text=True)
        visible_texts = filter(TableExtraction.tag_visible, texts)
        # print([x.strip() for x in visible_texts])
        # exit(0)
        return u" ".join(t.strip() for t in visible_texts if t.strip() != "")


class TableExtractor(Extractor):
    """
    **Description**
        Adding explanation here

    Examples:
        ::

            table_extractor = TableExtractor()
            table_extractor.extract(html=html_str,
                                    return_text=True)

    """

    def __init__(self) -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="content",
                           name="DigTableExtractor")
        self.tableExtractorInstance = TableExtraction()

    def _wrap_value_with_context(self, value: dict or str, field_name: str, start: int=0, end: int=0) -> Extraction:
        """Wraps the final result"""
        return Extraction(value, self.name, start_token=start, end_token=end, tag=field_name)

    def extract(self, html: str, return_text: bool = False, expand_merged_cells: bool = True) -> List[Extraction]:
        """
        Args:
            html (str): raw html of the page
            return_text (bool): if True, return the visible text in the page
                                removing all the data tables

        Returns:
            List[Extraction]: a list of Extractions

        """
        results = list()
        temp_res = self.tableExtractorInstance.extract(html, expand_merged_cells=expand_merged_cells)
        if return_text:
            results.append(self._wrap_value_with_context(temp_res['html_text'], "text_without_tables"))
        results.extend(map(lambda t: self._wrap_value_with_context(t, "tables"), temp_res['tables']))
        return results



row_key = "Rows"
column_key = "Cols"
special_keywords = ['jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov',
                    'dec', 'january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september',
                    'october', 'november', 'december', 'q1', 'q2', 'q3', 'q4']

# cell classes
text_cell = "text"
number_cell = "number"
empty_cell = "empty"
date_cell = "date"
blank_cell = "blank"

row_orientation = "row"
column_orientation = "col"


def data_to_string(data):
    return str(data)



from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from typing import List
from email.utils import parsedate_to_datetime
from bs4 import BeautifulSoup
import re
import json
import datetime

class MailmanExtractor(Extractor):
    def __init__(self, email_url: str, mailing_list_name: str, extractor_name: str) -> None:

        """
        Initialize the extractor, storing mailing list and message information
        Args:
            email_url: str
            mailing_list_name: str
            extractor_name: str

        Returns:
        """
        
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="build_in_extractor",
                           name=extractor_name)

        self.email_url = email_url
        self.mailing_list_name = mailing_list_name

    def old_format(self, content: BeautifulSoup) -> List[str]:

        """
        Extracts email message information if it uses the old Mailman format
        Args:
            content: BeautifulSoup

        Returns: List[str]
        """
        
        b = content.find('body')
        sender, date, nxt, rep_to = None, None, None, None
        strongs = b.findAll('strong', recursive=False)
        for s in strongs:
            field = str(s).split(">")[1].split("<")[0]
            if 'From' in field:
                sender = s.next_sibling.split("(")[0].strip()
            elif 'Date' in field:
                date_str = s.next_sibling.strip().replace("-","").replace("  "," ").strip()
                try:
                    date = parsedate_to_datetime(date_str).isoformat()[:19]
                except:
                    date = None
        sender = b.find('b').text if sender == None else sender
        sender = b.find('a').text if len(sender) == 0 else sender
        date = b.find('i').text[:19] if date == None else date

        try:
            nav = content.find('ul').findAll('li')
        except:
            nav = None
        if nav != None:
            for l in nav:
                s = l.text
                if 'Next in thread' in s:
                    nxt = '/'.join(self.email_url.split('/')[:-1]) + '/' + l.find('a')['href']
                    nxt = nxt[1:] if nxt[0] == '/' else nxt
                elif 'reply to' in s:
                    rep_to = '/'.join(self.email_url.split('/')[:-1]) + '/' + l.find('a')['href']
                    rep_to = rep_to[1:] if rep_to[0] == '/' else rep_to
        body = content.find('pre')
        body = body.text.strip() if body != None else None
        return [str(i) for i in [sender, date, body, nxt, rep_to]]

    def new_format(self, navbar: BeautifulSoup, content: BeautifulSoup) -> List[str]:

        """
        Extracts email message information if it uses the new Mailman format
        Args:
            content: BeautifulSoup

        Returns: List[str]
        """
        
        sender = content.find(id='from').text.split('via')[0][6:].strip()
        date_str = content.find(id='date').text.split(': ')[1].strip()
        date = parsedate_to_datetime(date_str).isoformat()[:19]
        body = content.find(id='body').text.strip()
        nxt, rep_to = None, None
        
        links = navbar.findAll('a')
        for l in links:
            if 'Next in thread' in str(l):
                nxt = '/'.join(self.email_url.split('/')[:-1]) + '/' + l['href']
                nxt = nxt[1:] if nxt[0] == '/' else nxt
            elif 'reply to' in str(l):
                rep_to = '/'.join(self.email_url.split('/')[:-1]) + '/' + l['href']
                rep_to = rep_to[1:] if rep_to[0] == '/' else rep_to
        return [str(i) for i in [sender, date, body, nxt, rep_to]]
    
    def extract(self, text: str) -> List[Extraction]:

        """
        Extracts and structures email message from UTF8-encoded text
        Args:
            text: str

        Returns: Extraction
        """
        
        content = BeautifulSoup(text, 'html5lib')
        subject = content.find('h1').text.strip()
        recip = self.mailing_list_name
        
        navbar = content.find(id='navbar')
        if navbar == None:
            info = self.old_format(content)
        else:
            info = self.new_format(navbar, content)
        for i in info[0:3]:
            if i == 'None':
                print('missed something important')
        sender = info[0]
        date   = info[1]
        body   = info[2]
        nxt    = info[3]
        rep_to = info[4]
        pub = 'SeeSat_Obs'
        dRec = datetime.datetime.now().isoformat()
        
        msg_obj = { 
            'url' : self.email_url,
            '@context' : {
                '@vocab' : 'schema.org'
            },
            'subject' : subject,
            'recip' : recip,
            'sender' : sender
        }
        if date != 'None':
            msg_obj['dateReceived'] = date
        if body != 'None':
            msg_obj['body'] = body
        if nxt != 'None':
            msg_obj['nxt'] = nxt
        if rep_to != 'None':
            msg_obj['replyToMessage'] = rep_to
        return Extraction(value=msg_obj, extractor_name=self.name)



import requests
from requests.auth import HTTPBasicAuth


def add_arguments(parser):
    parser.add_argument('-e', '--endpoint', action='store', type=str, dest='endpoint')
    parser.add_argument('-f', '--input', action='store', type=str, dest='file')
    parser.add_argument('--user', action='store', default=None, type=str, dest='user')
    parser.add_argument('--passwd', action='store', default=None, type=str, dest='passwd')


def run(args):
    print('Start uploading file {} to endpoint {}'.format(args.file, args.endpoint))
    headers = {
        'Content-Type': 'application/x-turtle',
    }
    response = requests.post(args.endpoint, data=open(args.file).read(), headers=headers,
                             auth=HTTPBasicAuth(args.user, args.passwd))

    print('Upload finished with status code: {}!'.format(response.status_code))
    if response.status_code // 100 != 2:
        print(response.content)




from etk.wikidata.truthy import TruthyUpdater


def add_arguments(parser):
    parser.add_argument('-e', '--endpoint', action='store', type=str, dest='endpoint')
    parser.add_argument('-c', '--creator', default='http://www.isi.edu/datamart', action='store', type=str,
                        dest='creator')
    parser.add_argument('--user', action='store', default=None, type=str, dest='user')
    parser.add_argument('--passwd', action='store', default=None, type=str, dest='passwd')
    parser.add_argument('--dry-run', action='store_true', default=False, dest='dryrun')


def run(args):
    print('Start deleting statements created by {} in endpoint {}'.format(args.creator, args.endpoint))
    if args.dryrun:
        print('Mode: Dryrun')
    tu = TruthyUpdater(args.endpoint, args.dryrun, args.user, args.passwd)
    creator = args.creator

    query = '''
      DELETE {
        ?statement a wikibase:BestRank .
        ?entity ?wdt ?value .
        ?entity ?wdtn ?normalValue .
      } WHERE {
        ?statement <http://www.isi.edu/etk/createdBy> <%s> ;
                   a wikibase:BestRank .
        ?entity ?p ?statement .
        ?statement ?ps ?value .
        ?entity ?wdt ?value .
        FILTER (strStarts(str(?p), "http://www.wikidata.org/prop/") && strStarts(str(?ps), "http://www.wikidata.org/prop/statement/") && strStarts(str(?wdt), "http://www.wikidata.org/prop/direct/"))
        OPTIONAL {
            ?entity ?wdtn ?normalValue .
            ?statement ?psn ?normalValue .
            FILTER (strStarts(str(?psn), "http://www.wikidata.org/prop/statement/value-normalized/") && strStarts(str(?wdtn), "http://www.wikidata.org/prop/direct-normalized/"))
        }
      }
    ''' %  creator
    tu.update(query)


    query = '''
      DELETE {
        ?statement ?p1 ?o .
        ?s ?p2 ?statement
      } WHERE {
        ?statement <http://www.isi.edu/etk/createdBy> <%s> .
        ?statement ?p1 ?o 
        OPTIONAL { ?s ?p2 ?statement }
      }
    ''' % creator
    tu.update(query)

    query = '''
      DELETE {
        ?entity ?p1 ?o .
        ?s ?p2 ?entity
      } WHERE {
        { ?entity a wikibase:Item } UNION { ?entity a wikibase:Property }
        FILTER NOT EXISTS { ?entity ?p [ wikibase:rank ?rank ] }
        ?entity ?p1 ?o .
        OPTIONAL { ?s ?p2 ?entity }
      }
    '''
    tu.update(query)
    print('Deletion complete!')



import unittest
import json
from etk.extractors.html_metadata_extractor import HTMLMetadataExtractor


class TestMetadataExtractor(unittest.TestCase):
    # value of title extraction is a string
    def test_title_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html.jl', 'r') as f:
            sample_html = json.load(f)
        test_html = sample_html["raw_content"]

        title_text = hme.extract(test_html, extract_title=True)[0].value
        expected_title = "323-452-2013 ESCORT ALERT! - Luna The Hot Playmate (323) 452-2013 - 23"
        self.assertEqual(title_text, expected_title)

    # value of mate extraction is a dict
    def test_meta_tag_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html.jl', 'r') as f:
            sample_html = json.load(f)
        test_html = sample_html["raw_content"]

        meta_tag_dic = hme.extract(test_html, extract_meta=True)[0].value
        expected_dic = {
            'description': "323-452-2013 Escort listings. View all of her listings at once. On this page you will see a history for 323-452-2013 and all of the escort ads placed with this phone number. Specifically in losangeles: 323-452-2013 ESCORT ALERT! - Luna The Hot Playmate (323) 452-2013 - 23",
            'netinsert': "0.0.1.13.13.1",
            'keywords': "323-452-2013, escort listing, escort ad, escorts in Los angeles, 323-452-2013 in Los angeles, escorts in losangeles, call 323-452-2013, phone number 323-452-2013, craigslist escorts, backpage escorts, listings"
        }

        self.assertDictEqual(meta_tag_dic, expected_dic)

    # value of microdata extraction is a list
    def test_microdata_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        test_html = """<!DOCTYPE HTML>
                    ... <html>
                    ...  <head>
                    ...   <title>Photo gallery</title>
                    ...  </head>
                    ...  <body>
                    ...   <h1>My photos</h1>
                    ...   <figure itemscope itemtype="http://n.whatwg.org/work" itemref="licenses">
                    ...    <img itemprop="work" src="images/house.jpeg" alt="A white house, boarded up, sits in a forest.">
                    ...    <figcaption itemprop="title">The house I found.</figcaption>
                    ...   </figure>
                    ...   <figure itemscope itemtype="http://n.whatwg.org/work" itemref="licenses">
                    ...    <img itemprop="work" src="images/mailbox.jpeg" alt="Outside the house is a mailbox. It has a leaflet inside.">
                    ...    <figcaption itemprop="title">The mailbox.</figcaption>
                    ...   </figure>
                    ...   <footer>
                    ...    <p id="licenses">All images licensed under the <a itemprop="license"
                    ...    href="http://www.opensource.org/licenses/mit-license.php">MIT
                    ...    license</a>.</p>
                    ...   </footer>
                    ...  </body>
                    ... </html>"""

        microdata = hme.extract(test_html, extract_microdata=True)[0].value
        expected_microdata = [{"type": "http://n.whatwg.org/work",
                               "properties": {"license": "http://www.opensource.org/licenses/mit-license.php",
                                              "work": "images/house.jpeg",
                                              "title": "The house I found."}}, {"type": "http://n.whatwg.org/work",
                                        "properties": {"license": "http://www.opensource.org/licenses/mit-license.php",
                            "work": "images/mailbox.jpeg", "title": "The mailbox."}}]
        self.assertEqual(microdata, expected_microdata)

    # value of microdata extraction is list
    def test_json_ld_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html_for_json_ld.html', 'r') as f:
            test_html = f.read()

        json_ld_data = hme.extract(test_html, extract_json_ld=True)[0].value

        expected_ld_data = [{'@context': 'http://schema.org',
                             '@type': 'MusicEvent',
                             'location': {'@type': 'Place',
                                          'address': {'@type': 'PostalAddress',
                                                      'addressCountry': 'US',
                                                      'addressLocality': 'Brooklyn',
                                                      'addressRegion': 'NY',
                                                      'postalCode': '11225',
                                                      'streetAddress': '497 Rogers Ave'},
                                          'geo': {'@type': 'GeoCoordinates',
                                                  'latitude': 40.660109,
                                                  'longitude': -73.953193},
                                          'name': 'The Owl Music Parlor',
                                          'sameAs': 'http://www.theowl.nyc'},
                             'name': 'Elysian Fields',
                             'performer': [{'@type': 'MusicGroup',
                                            'name': 'Elysian Fields',
                                            'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'}],
                             'startDate': '2017-06-10T19:30:00-0400',
                             'url': 'https://www.songkick.com/concerts/30173984-elysian-fields-at-owl-music-parlor?utm_medium=organic&utm_source=microformat'},
                            {'@context': 'http://schema.org',
                             '@type': 'MusicEvent',
                             'location': {'@type': 'Place',
                                          'address': {'@type': 'PostalAddress',
                                                      'addressCountry': 'US',
                                                      'addressLocality': 'San Francisco',
                                                      'addressRegion': 'CA',
                                                      'postalCode': '94107',
                                                      'streetAddress': '500 Fourth Street'},
                                          'geo': {'@type': 'GeoCoordinates',
                                                  'latitude': 37.7795638,
                                                  'longitude': -122.398023},
                                          'name': 'Hotel Utah Saloon',
                                          'sameAs': 'http://www.hotelutah.com/'},
                             'name': 'Elysian Fields',
                             'performer': [{'@type': 'MusicGroup',
                                            'name': 'Elysian Fields',
                                            'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'},
                                           {'@type': 'MusicGroup',
                                            'name': 'Chocolate Genius Inc.',
                                            'sameAs': 'https://www.songkick.com/artists/1009602-chocolate-genius-inc?utm_medium=organic&utm_source=microformat'}],
                             'startDate': '2017-04-26T20:00:00-0700',
                             'url': 'https://www.songkick.com/concerts/29673614-elysian-fields-at-hotel-utah-saloon?utm_medium=organic&utm_source=microformat'},
                            {'@context': 'http://schema.org',
                             '@type': 'MusicEvent',
                             'location': {'@type': 'Place',
                                          'address': {'@type': 'PostalAddress',
                                                      'addressCountry': 'France',
                                                      'addressLocality': 'Saint-Nazaire',
                                                      'postalCode': '44600',
                                                      'streetAddress': 'Alvéole 14 de la base '
                                                                       'sous-Marine Bd de la '
                                                                       'Légion d’Honneur'},
                                          'geo': {'@type': 'GeoCoordinates',
                                                  'latitude': 47.2755434,
                                                  'longitude': -2.2022817},
                                          'name': 'VIP',
                                          'sameAs': 'http://www.levip-saintnazaire.com/'},
                             'name': 'Elysian Fields',
                             'performer': [{'@type': 'MusicGroup',
                                            'name': 'Elysian Fields',
                                            'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'},
                                           {'@type': 'MusicGroup',
                                            'name': 'Troy Von Balthazar',
                                            'sameAs': 'https://www.songkick.com/artists/355304-troy-von-balthazar?utm_medium=organic&utm_source=microformat'}],
                             'startDate': '2016-10-29T21:00:00+0200',
                             'url': 'https://www.songkick.com/concerts/27626524-elysian-fields-at-vip?utm_medium=organic&utm_source=microformat'},
                            {'@context': 'http://schema.org',
                             '@type': 'MusicGroup',
                             'image': 'https://images.sk-static.com/images/media/profile_images/artists/236156/card_avatar',
                             'interactionCount': '6066 UserLikes',
                             'logo': 'https://images.sk-static.com/images/media/profile_images/artists/236156/card_avatar',
                             'name': 'Elysian Fields',
                             'url': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'}]
        self.assertEqual(json_ld_data, expected_ld_data)

    # value of microdata extraction is a list
    def test_rdfa_extrator(self) -> None:
        hme = HTMLMetadataExtractor()
        test_html = """<html>
                    ...  <head>
                    ...    ...
                    ...  </head>
                    ...  <body prefix="dc: http://purl.org/dc/terms/ schema: http://schema.org/">
                    ...    <div resource="/alice/posts/trouble_with_bob" typeof="schema:BlogPosting">
                    ...       <h2 property="dc:title">The trouble with Bob</h2>
                    ...       <h3 property="dc:creator schema:creator" resource="#me">Alice</h3>
                    ...       <div property="schema:articleBody">
                    ...         <p>The trouble with Bob is that he takes much better photos than I do:</p>
                    ...       </div>
                    ...    </div>
                    ...  </body>
                    ... </html>
                    ... """

        rdfa_data = hme.extract(test_html, extract_rdfa=True)[0].value
        expected_rdfa_data = [{'@id': '/alice/posts/trouble_with_bob',
                               '@type': ['http://schema.org/BlogPosting'], 'http://purl.org/dc/terms/creator':
                                   [{'@id': '#me'}], 'http://purl.org/dc/terms/title':
                                   [{'@value': 'The trouble with Bob'}], 'http://schema.org/articleBody':
                                   [{'@value': '\n                    ...         '
                                               'The trouble with Bob is that he takes much better photos than I do:\n   '
                                               '                 ...       '}],
                               'http://schema.org/creator': [{'@id': '#me'}]}]
        self.assertEqual(rdfa_data, expected_rdfa_data)


if __name__ == '__main__':
    unittest.main()



from etk.timeseries.annotation.table_processor import parsed_table
from etk.timeseries.annotation.utility import column_orientation
from etk.timeseries.annotation import date_utilities


class complex_table(parsed_table):
    # finding the overlapp in the time block based on monotonicity of the time block
    @classmethod
    def check_for_overlapping_time(cls, date_block, orientation, raw_sheet):
        check_points = []
        is_increasing = True
        if orientation == column_orientation:  # times are in columns
            check_points.append(date_block.upper_row)
            if date_block.lower_row - date_block.upper_row >= 2:
                d = date_utilities.parse_row_as_date(
                    raw_sheet[date_block.upper_row][date_block.left_col:date_block.right_col])
                d2 = date_utilities.parse_row_as_date(
                    raw_sheet[date_block.upper_row + 1][date_block.left_col:date_block.right_col])
                is_increasing = date_utilities.is_greater(d, d2)
            for row in range(date_block.upper_row, date_block.lower_row - 1):
                d = date_utilities.parse_row_as_date(raw_sheet[row][date_block.left_col:date_block.right_col])

                d2 = date_utilities.parse_row_as_date(raw_sheet[row + 1][date_block.left_col:date_block.right_col])
                if not date_utilities.is_greater(d, d2) and is_increasing:
                    check_points.append(row + 1)
                if not date_utilities.is_greater(d2, d) and not is_increasing:
                    check_points.append(row + 1)
            check_points.append(date_block.lower_row)
        else:
            check_points.append(date_block.left_col)
            check_points.append(date_block.right_col)
        # ToDO: write the same process for the row_orientation
        if len(check_points) == 2:
            return False, check_points

        return True, check_points



from etk.timeseries.simple_annotator import SimpleAnnotator
from etk.timeseries.extract_spreadsheet import ExtractSpreadsheet


class SimpleTimeSeriesExtractor(object):
    """
    This class will take a "simple" csv or excel file as input and out a timeseries, if there is a timeseries to be
    extracted.
    """

    def __init__(self, csv_file_path: str):
        self.csv_file_path = csv_file_path

    def process(self):
        simple_annotator = SimpleAnnotator(self.csv_file_path)
        # returns the annotation object for the csv
        annotations = simple_annotator.get_annotation_json()
        es = ExtractSpreadsheet(self.csv_file_path, annotations=annotations)
        return list(es.process())


# if __name__ == '__main__':
#     import json
#     ste = SimpleTimeSeriesExtractor('/Users/amandeep/Github/sage-research-tool/datasets/fred_US_AL/example/DEXUSAL.csv')
#     print(json.dumps(ste.process(), indent=2))



import unittest
from etk.knowledge_graph.graph import Graph
from etk.knowledge_graph.schema import KGSchema
from etk.knowledge_graph.subject import Subject
from etk.knowledge_graph.node import URI, BNode, Literal, LiteralType


MASTER_CONFIG_EXAMPLE = '''
{
    "fields": {
        "cameo_code": {
            "type": "string", 
            "description": "CAMEO code to describe the type of event in the ICEWS dataset"
        }, 
        "answer_option": {
            "type": "kg_id" 
        }, 
        "reference": {
            "type": "string"
        }
    }
}
'''

Ontology_EXAMPLE = '''
@prefix : <http://dig.isi.edu/ontologies/dig/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
:Entity a owl:Class ;
    rdfs:label "Entity" ;
    skos:definition """Everything that can be described, similar to owl:Thing.""" ;
    :common_properties :label, :title, :description ; .
'''


class TestKGSchema(unittest.TestCase):
    def test_kg_schema(self):
        schema = KGSchema()
        schema.add_schema(Ontology_EXAMPLE, 'ttl')

    def test_kg_schema_master_config(self):
        schema = KGSchema()
        schema.add_schema(MASTER_CONFIG_EXAMPLE, 'master_config')
        self.assertEqual(len(schema.fields), 3)



from typing import List, Dict
from dateutil import parser
from datetime import timedelta
import datetime
from etk.extractors.date_extractor import DateExtractor, DateResolution
import sys

de = DateExtractor(extractor_name='date_extractor')

"""
Detect simple granularity types in a sequence of dates.
"""


class GranularityDetector:
    granularity = {
        "second": timedelta(seconds=1),
        "minute": timedelta(minutes=1),
        "hour": timedelta(hours=1),
        "day": timedelta(hours=24),
        "week": timedelta(days=7),
        "month": timedelta(days=30),
        "quarter": timedelta(days=120),
        "year": timedelta(days=365)
    }

    @staticmethod
    def get_parsed_date(my_date):
        if isinstance(my_date, datetime.datetime):
            return my_date
        elif isinstance(my_date, datetime.date):
            return datetime.datetime(my_date.year, my_date.month, my_date.day)
        elif isinstance(my_date, str):
            try:
                extraction = de.extract(my_date,
                                        date_value_resolution=DateResolution.SECOND,
                                        use_default_formats=True,
                                        additional_formats=['%Y'])
                iso_date = extraction[0].value
                parsed_date = parser.parse(iso_date) if iso_date else None

                return parsed_date
            except:
                return None
        return None

    @staticmethod
    def get_granularity(dates: List, return_best=True, min_confidence=0.5) -> Dict or str:
        try:
            if len(dates) == 1:
                return "unknown"

            granularity_confidence = dict()
            parsed_dates = [GranularityDetector.get_parsed_date(date) for date in dates]

            for g in GranularityDetector.granularity:
                time_delta = GranularityDetector.granularity[g]
                confidence = 0.0

                for i in range(1, len(parsed_dates)):
                    if parsed_dates[i] and parsed_dates[i - 1]:
                        time_diff = abs(parsed_dates[i] - parsed_dates[i - 1])
                        error = abs(time_diff - time_delta) / (
                                    time_diff + time_delta)  # Is there a better measure for error?
                        confidence += 1 - error

                confidence /= (len(parsed_dates) - 1)
                granularity_confidence[g] = confidence

            if return_best:
                most_confident = max(granularity_confidence.keys(), key=lambda key: granularity_confidence[key])
                if granularity_confidence[most_confident] >= min_confidence:
                    return most_confident
                else:
                    return "unknown"
            else:
                return granularity_confidence

        except:
            print("Unexpected error:", sys.exc_info()[0])

            if return_best:
                return "unknown"
            else:
                return dict()



import unittest
from etk.knowledge_graph.node import URI, BNode, Literal
from etk.knowledge_graph.subject import Subject, Reification, InvalidParameter


class TestKGTriples(unittest.TestCase):
    def test_subject(self):
        s = URI('ex:ex1')
        t = Subject(s)
        lit = Literal('jack', 'en', 'xsd:string')
        t.add_property(URI('rdf:type'), URI('dig:Person'))
        t.add_property(URI('dig:name'), lit)
        self.assertEqual(t.subject, s)
        self.assertEqual(len(t._resource), 2)
        self.assertIn(URI('rdf:type'), t._resource)
        self.assertIn(URI('dig:name'), t._resource)

        # remove property
        self.assertFalse(t.remove_property(URI('rdf:type'), URI('dig:Male')))
        self.assertTrue(t.remove_property(URI('rdf:type'), URI('dig:Person')))
        self.assertNotIn(URI('rdf:type'), t._resource)

        self.assertTrue(t.remove_property(URI('dig:name')))
        self.assertNotIn(URI('dig:name'), t._resource)

        self.assertFalse(t.remove_property(URI('rdf:type')))
        self.assertFalse(t.remove_property(URI('dig:name')))

    def test_subject_exception(self):
        with self.assertRaises(InvalidParameter):
            Subject(None)

        with self.assertRaises(InvalidParameter):
            Subject('ex:ex1')

        with self.assertRaises(InvalidParameter):
            Subject(Literal('test'))

        s = Subject(URI('ex:ex1'))
        with self.assertRaises(InvalidParameter):
            s.add_property(URI('rdf:type'), 'dig:Person')

        with self.assertRaises(InvalidParameter):
            s.add_property(BNode(), URI('dig:Person'))

    def test_subject_reification(self):
        s = URI('ex:ex1')
        t = Subject(s)
        lit = Literal('jack', 'en', 'xsd:string')
        statement1 = t.add_property(URI('rdf:type'), URI('dig:Person'))
        statement2 = t.add_property(URI('dig:name'), lit, Reification(URI('digg:name')))
        statement3 = t.add_property(URI('dig:name'), lit, Reification(URI('diga:name'), URI('digb:name')))
        statement4 = t.add_property(URI('dig:name'), lit, Reification(URI('digg:name'), statement=URI('ex:assert1')))
        self.assertEqual(t.subject, s)
        self.assertEqual(len(t._resource), 4)
        self.assertIn(URI('rdf:type'), t._resource)
        self.assertIn(URI('dig:name'), t._resource)
        self.assertIn(URI('digg:name'), t._resource)
        self.assertIn(URI('diga:name'), t._resource)
        self.assertNotIn(URI('digb:name'), t._resource)
        self.assertNotIsInstance(statement1, Subject)
        self.assertIsInstance(statement2, Subject)
        self.assertIn(URI('digg:name'), statement2._resource)
        self.assertIsInstance(statement3, Subject)
        self.assertIsInstance(statement3.subject, BNode)
        self.assertIsInstance(statement4, Subject)
        self.assertIsInstance(statement4.subject, URI)

        # exception
        with self.assertRaises(InvalidParameter):
            Reification(None)
        with self.assertRaises(InvalidParameter):
            Reification(URI('dig:name'), 3)
        with self.assertRaises(InvalidParameter):
            Reification(URI('dig:name'), statement='')

    def test_subject_reification_back_compatible(self):
        s = URI('ex:ex1')
        t = Subject(s)
        lit = Literal('jack', 'en', 'xsd:string')
        statement1 = t.add_property(URI('rdf:type'), URI('dig:Person'))
        statement2 = t.add_property(URI('dig:name'), lit, reify=URI('digg:name'))
        statement3 = t.add_property(URI('dig:name'), lit, reify=(URI('digg:name'), None))
        statement4 = t.add_property(URI('dig:name'), lit, reify=(URI('digg:name'), URI('ex:assert1')))
        self.assertEqual(t.subject, s)
        self.assertEqual(len(t._resource), 3)
        self.assertIn(URI('rdf:type'), t._resource)
        self.assertIn(URI('dig:name'), t._resource)
        self.assertIn(URI('digg:name'), t._resource)
        self.assertNotIsInstance(statement1, Subject)
        self.assertIsInstance(statement2, Subject)
        self.assertIn(URI('digg:name'), statement2._resource)
        self.assertIsInstance(statement3, Subject)
        self.assertIsInstance(statement3.subject, BNode)
        self.assertIsInstance(statement4, Subject)
        self.assertIsInstance(statement4.subject, URI)

        # exception
        with self.assertRaises(InvalidParameter):
            t.add_property(URI('dig:name'), lit, reify=(None, URI('s')))
        with self.assertRaises(InvalidParameter):
            t.add_property(URI('dig:name'), lit, reify=(URI('digg:name'), 3))
        with self.assertRaises(InvalidParameter):
            t.add_property(URI('dig:name'), lit, reify=(URI('digg:name'), ''))




import unittest
from etk.knowledge_graph.graph import Graph
from etk.knowledge_graph.subject import Subject
from etk.knowledge_graph.node import URI, BNode, Literal, LiteralType


class TestKGGraph(unittest.TestCase):
    def test_graph(self):
        g = Graph()
        g.bind('ex', 'http://ex.com/')
        g.bind('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        g.bind('foaf', 'http://xmlns.com/foaf/0.1/')

        t_nest = Subject(BNode())
        t_nest.add_property(URI('rdf:type'), URI('foaf:Person'))
        t_nest.add_property(URI('foaf:name'), Literal('Alfred'))
        t_nest.add_property(URI('foaf:age'), Literal('36', type_=LiteralType.int))

        t = Subject(URI('ex:john'))
        t.add_property(URI('rdf:type'), URI('foaf:Person'))
        t.add_property(URI('foaf:name'), Literal('John'))
        t.add_property(URI('foaf:age'), Literal('12', type_=LiteralType.int))
        t.add_property(URI('foaf:knows'), t_nest)
        g.add_subject(t)

        self.assertEqual(len(g._g), 7)
        # print(g.serialize())
        # print('========================================================================')
        # print(g.serialize('nt'))
        # print('========================================================================')
        # context = {'@vocab': 'http://xmlns.com/foaf/0.1/'}
        # print(g.serialize('json-ld', context=context, auto_compact=True))



import datetime
import logging
import re

months = [ 'jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov','dec', 'january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september',
            'october', 'november', 'december']
exact_time = {"jan":1, "feb":2, "mar":3, "apr":4, "may":5, "jun":6, "jul":7, "aug":8, "sep":9, "oct":10, "nov":11, "dec":12,
            "january":1, "february":2, "march":3, "april":4, "june":6, "july":7, "august":8, "september":9, "october":10, "november":11, "december":12,
              }

special_keywords = [ 'jan', 'feb', 'mar', 'apr', 'may', 'jun', 'jul', 'aug', 'sep', 'oct', 'nov',
                    'dec', 'january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september',
                    'october', 'november', 'december', 'q1', 'q2', 'q3', 'q4']

date_parts = {"%Y":'year', "%B":'month', "%d": 'day'}

# a date point
class parsed_date:
    def __init__(self):
        self.is_range = False
        self.predicted_date = dict()

    def get_printable_version(self):
        return self.predicted_date

    def equals(self, p):
        for i in self.predicted_date.keys():
            if i not in p.predicted_date.keys():
                return False
            if p.predicted_date[i] != self.predicted_date[i]:
                return False
        return True

class parsed_interval:
    def __init__(self):
        self.begin_interval = parsed_date()
        self.end_interval = parsed_date()
        self.predicted_date_certainty = dict()

    def equals(self, p):
        if not self.begin_interval.equals(p.begin_interval):
            return False
        if not self.end_interval.equals(p.end_interval):
            return False
        return True
    def print_interval(self):
        stres = "begin: " + str(self.begin_interval.get_printable_version()) +  " end: " + str(self.end_interval.get_printable_version())
        print(stres)

# direction = True means the first element in year, direction False means the last element is last
def parse_pattern(cell, pattern , direction, day_available):
    guessed_date = parsed_interval()
    matchObj = re.match(pattern, cell, re.M | re.I)

    dir = [1, 2, 3]
    guessed_date.begin_interval.predicted_date['year'] = int(matchObj.group(dir[0]))
    guessed_date.end_interval.predicted_date['year'] = int(matchObj.group(dir[0]))
    guessed_date.predicted_date_certainty['year'] = 1
    guessed_date.begin_interval.predicted_date['month'] = int(matchObj.group(dir[1]))
    guessed_date.end_interval.predicted_date['month'] = int(matchObj.group(dir[1]))
    guessed_date.predicted_date_certainty['month'] = 1


    if day_available:
        guessed_date.begin_interval.predicted_date['day'] = int(matchObj.group(3))
        guessed_date.end_interval.predicted_date['day'] = int(matchObj.group(3))
        guessed_date.predicted_date_certainty['day'] = 1
    else:
        guessed_date.begin_interval.predicted_date['day'] = 1
        guessed_date.end_interval.predicted_date['day'] = 30
        guessed_date.predicted_date_certainty['day'] = 0.5


def parse_date(cell):
    guessed_date = parsed_interval()
    if type(cell) == datetime.date or type(cell) == datetime.datetime:
        guessed_date.begin_interval.predicted_date[date_parts["%Y"]]= int(cell.strftime("%Y")) # this is the year
        guessed_date.end_interval.predicted_date[date_parts["%Y"]] = int(cell.strftime("%Y"))
        guessed_date.predicted_date_certainty[date_parts["%Y"]] = 1

        guessed_date.begin_interval.predicted_date[date_parts["%B"]] = int(exact_time[cell.strftime("%B").lower()])
        guessed_date.end_interval.predicted_date[date_parts["%B"]] = int(exact_time[cell.strftime("%B").lower()])
        guessed_date.predicted_date_certainty[date_parts["%B"]] = 1


        guessed_date.begin_interval.predicted_date[date_parts["%d"]] = int(cell.strftime("%d"))
        guessed_date.end_interval.predicted_date[date_parts["%d"]] = int(cell.strftime("%d"))
        if int(cell.strftime("%d")) == 1: # not working dont know why
            guessed_date.predicted_date_certainty[date_parts["%d"]] = 0.5
            guessed_date.end_interval.predicted_date[date_parts["%d"]] = 30

            if exact_time[cell.strftime("%B").lower()] == 1:  # if it was not numerical
                guessed_date.predicted_date_certainty[date_parts["%B"]] = 0.5
                guessed_date.end_interval.predicted_date[date_parts["%B"]] = 12

        else:
            guessed_date.predicted_date_certainty[date_parts["%d"]] = 1
        #logging.info("")
             # this is day -> what if the day is not indicated? maybe this is not valid... I should have a bit vector specially when it has the value exactly "1"
        return guessed_date
    # if the type is date and it is a text:
    if type(cell) == int:
        if cell > 1900 and cell < 2020:
            guessed_date.begin_interval.predicted_date['year'] = cell
            guessed_date.end_interval.predicted_date['year'] = cell
            guessed_date.predicted_date_certainty['year'] = 1
    else:

        try:
            # print cell
            if int(cell) > 1900 and int(cell) < 2020: #current year should be added
                guessed_date.begin_interval.predicted_date['year'] = int(cell)
                guessed_date.end_interval.predicted_date['year'] = int(cell)
                guessed_date.predicted_date_certainty['year'] = 1
        except:
            if cell.lower() in exact_time.keys():
                guessed_date.begin_interval.predicted_date['month'] = exact_time[cell.lower()]
                guessed_date.end_interval.predicted_date['month'] = exact_time[cell.lower()]
                guessed_date.predicted_date_certainty['month'] = 1
                return guessed_date
            elif re.match('q\d', cell.lower()):
                guessed_date.begin_interval.predicted_date['month'] = (int(cell[1])-1) * 3 + 1
                guessed_date.begin_interval.predicted_date['day'] = 1
                guessed_date.end_interval.predicted_date['month'] = int(cell[1])*3
                guessed_date.end_interval.predicted_date['day'] = 30
                guessed_date.predicted_date_certainty['month'] = 1
                guessed_date.predicted_date_certainty['day'] = 0.5
            elif len(cell.lower().split('-')) > 1:
                # patterns = [r'(\d+)-(\d+)', r'(\d+)-(\d+)-(\d+)']
                #
                # for x in patterns:
                #      matchObj = re.match(x, cell, re.M | re.I)
                #      if matchObj:
                #          if int(matchObj.group(1)) > 1000:  # this is perhaps the year
                #              if len(cell.lower.split('-')) > 2:
                #                  return parse_pattern(cell, x, True, True)
                #              else:
                #                  return parse_pattern(cell, x, True, False)

                     # complete for the reverse direction

                sub_dates = cell.lower().split('-') # it can be a range or it can be a specific date
                guessed_date.begin_interval = parse_date(sub_dates[0]).begin_interval
                guessed_date.predicted_date_certainty = parse_date(sub_dates[0]).predicted_date_certainty
                guessed_date.end_interval = parse_date(sub_dates[1]).begin_interval
            elif len(cell.lower().split(' ')) > 1:
                sub_dates = cell.lower().split(' ')  # it can be a range or it can be a specific date
                guessed_date.begin_interval = parse_date(sub_dates[0]).begin_interval
                guessed_date.predicted_date_certainty = parse_date(sub_dates[0]).predicted_date_certainty
                guessed_date.end_interval = parse_date(sub_dates[1]).begin_interval
            elif re.match(r'(\d+)[^\d+](\d+)', cell, re.M | re.I):
                matchObj = re.match(r'(\d+)[^\d+](\d+)', cell)
                guessed_date.begin_interval.predicted_date['year'] = int(matchObj.group(1))
                guessed_date.end_interval.predicted_date['year'] = int(matchObj.group(1))
                guessed_date.predicted_date_certainty['year'] = 1
                guessed_date.begin_interval.predicted_date['month'] = int(matchObj.group(2))
                guessed_date.end_interval.predicted_date['month'] = int(matchObj.group(2))
                guessed_date.predicted_date_certainty['month'] = 1

            else:
                logging.error("could not parse this date " + str(cell))
    return guessed_date

# we assume that currently we have daily monthly and yearly --> (quarterly and weekly will be added later
def find_date_intervals_differences(date_range1, date_range2):
    if 'year' in date_range1.begin_interval.predicted_date.keys() and 'year' in date_range2.begin_interval.predicted_date.keys():
        if date_range2.begin_interval.predicted_date['year'] != date_range1.begin_interval.predicted_date['year']:
            return 'yearly'

    if 'month' in date_range1.begin_interval.predicted_date.keys() and 'month' in date_range2.begin_interval.predicted_date.keys():
        if date_range2.begin_interval.predicted_date['month'] != date_range1.begin_interval.predicted_date['month']:
            if abs(date_range2.begin_interval.predicted_date['month'] - date_range1.begin_interval.predicted_date['month']) >= 3:
                return 'quarterly'
            else:
            #    print 'strange' + abs(date_range2.begin_interval.predicted_date['month'] - date_range1.begin_interval.predicted_date['month'])
                return 'monthly' # or maybe quarterly

    if 'day' in date_range1.begin_interval.predicted_date.keys() and 'day' in date_range2.begin_interval.predicted_date.keys():
        if date_range2.begin_interval.predicted_date['day'] != date_range1.begin_interval.predicted_date['day']:
            return 'daily'  # or maybe quarterly

    # add other granularities

    return 'no idea'

# parse the whole row as a unique timee interval
def parse_row_as_date(row):
    if len(row) == 1:
        return parse_date(row[0])

    # should start and shrink the date
    predicted_date = parse_date(row[0])
    for i in range(len(row) - 1):
        pr_date = parse_date(row[i+1])
        for j in pr_date.begin_interval.predicted_date.keys():
            if pr_date.predicted_date_certainty[j] == 1:
                # check the overlapp of two time intervals
                if j in predicted_date.begin_interval.predicted_date.keys():
                    if predicted_date.predicted_date_certainty[j] == 1:
                    # compare the intervals here
                        if predicted_date.begin_interval.predicted_date[j] >= pr_date.begin_interval.predicted_date[j] and predicted_date.end_interval.predicted_date[j] <= pr_date.end_interval.predicted_date[j]:
                            continue
                        # if it is finer then shrink the interval here
                        elif predicted_date.begin_interval.predicted_date[j] <= pr_date.begin_interval.predicted_date[j] and predicted_date.end_interval.predicted_date[j] >= pr_date.end_interval.predicted_date[j] :
                            predicted_date.begin_interval.predicted_date[j] = pr_date.begin_interval.predicted_date[j]
                            predicted_date.end_interval.predicted_date[j] = pr_date.end_interval.predicted_date[j]
                            predicted_date.predicted_date_certainty[j] = 1
                        continue
                else:
                    predicted_date.begin_interval.predicted_date[j] = pr_date.begin_interval.predicted_date[j]
                    predicted_date.end_interval.predicted_date[j] = pr_date.end_interval.predicted_date[j]
                    predicted_date.predicted_date_certainty[j] = pr_date.predicted_date_certainty[j]

    return predicted_date

def find_granularity(rows, left_lim, right_lim):
    founded_gr = {}
    for i in range(len(rows) - 1):
        founded_gr[find_date_intervals_differences(parse_row_as_date(rows[i][left_lim:right_lim]), parse_row_as_date(rows[i + 1][left_lim:right_lim]))]=1
    # print founded_gr
    # return the finest granularity
    if 'daily' in founded_gr:
        return 'daily'
    if 'monthly' in founded_gr:
        return 'monthly'
    if 'quarterly' in founded_gr:
        return 'quarterly'
    if 'yearly' in founded_gr:
        return 'yearly'

    return 'No idea'

#check if the given row can be a real date interval(no conflict between the cells)
def valid_time_row(row):
    predicted_date = parse_date(row[0])
    if len(row) == 1:
        return True, predicted_date

    for i in range(len(row) - 1):
        pr_date = parse_date(row[i+1])
        for j in pr_date.begin_interval.predicted_date.keys():
            if pr_date.predicted_date_certainty[j] == 1:
                # check the overlapp of two time intervals
                if j in predicted_date.begin_interval.predicted_date.keys():
                    if predicted_date.predicted_date_certainty[j] == 1:
                        if predicted_date.begin_interval.predicted_date[j] >= pr_date.begin_interval.predicted_date[j] and predicted_date.end_interval.predicted_date[j] <= pr_date.end_interval.predicted_date[j]:
                            continue
                        # if it is finer then shrink the interval here
                        elif predicted_date.begin_interval.predicted_date[j] <= pr_date.begin_interval.predicted_date[j] and predicted_date.end_interval.predicted_date[j] >= pr_date.end_interval.predicted_date[j] :
                            predicted_date.begin_interval.predicted_date[j] = pr_date.begin_interval.predicted_date[j]
                            predicted_date.end_interval.predicted_date[j] = pr_date.end_interval.predicted_date[j]
                            predicted_date.predicted_date_certainty[j] = 1
                            continue
                        return False, None
                else:
                    predicted_date.begin_interval.predicted_date[j] = pr_date.begin_interval.predicted_date[j]
                    predicted_date.end_interval.predicted_date[j] = pr_date.end_interval.predicted_date[j]
                    predicted_date.predicted_date_certainty[j] = pr_date.predicted_date_certainty[j]

    return True, pr_date

def is_greater(date1, date2):
    key_w = ['year', 'month', 'day']
    for w in key_w:
        if w in date1.end_interval.predicted_date.keys() and w in date2.begin_interval.predicted_date.keys() and date1.predicted_date_certainty[w] == 1 and date2.predicted_date_certainty[w] == 1:
            if date2.begin_interval.predicted_date[w] < date1.end_interval.predicted_date[w]:
                return False
            elif date2.begin_interval.predicted_date[w] > date1.end_interval.predicted_date[w]:
                return True
    return True


from etk.timeseries.annotation import block_detector
from etk.timeseries.annotation import date_utilities
from etk.timeseries.annotation import utility
import logging


class parsed_table:
    # we defind a table when we find a time header for that
    def __init__(self, time_header, parent_sheet):
        self.time_block = time_header
        self.parent_sheet = parent_sheet
        self.time_orientation = self.find_orientation()
        self.find_series_block()  # location of time series data
        self.labels = []
        self.label_names = {}
        self.label_backfill_bool = {}
        self.borders = None
        self.granularity = None
        self.find_granularity()
        self.back_mode_time = False
        self.break_points = []
        self.time_block_length = self.find_time_block_length()
        self.offset = 0

    # check if there exists consecutive rows in the time block which are parsed to a unique time
    def has_equivalent_time_rows(self):
        for i in range(self.time_block.upper_row, self.time_block.lower_row-1):
            p_date1 = date_utilities.parse_row_as_date(self.parent_sheet.raw_values[i][self.time_block.left_col:self.time_block.right_col])
            p_date2 = date_utilities.parse_row_as_date(self.parent_sheet.raw_values[i+1][self.time_block.left_col:self.time_block.right_col])
            if p_date1.equals(p_date2):
                return True
        return False

    def has_equivalent_time_columns(self):
        for i in range(self.time_block.left_col, self.time_block.right_col-1):
            tb = (list(self.parent_sheet.raw_values.columns()))
            p_date1 = date_utilities.parse_row_as_date(list(self.parent_sheet.raw_values.columns())[i][self.time_block.upper_row:self.time_block.lower_row])
            p_date2 = date_utilities.parse_row_as_date(list(self.parent_sheet.raw_values.columns())[i+1][self.time_block.upper_row:self.time_block.lower_row])
            if p_date1.equals(p_date2):
                return True
        return False

    ##important: using this offset list for granularity would be a good idea. specially when they are irregular
    # also it can be used for them to create annotation for files with irregular merged cells
    def has_regular_offset(self):
        offsets = []
        if self.time_orientation == utility.row_orientation:
            current_pattern = 0
            for i in range(self.time_block.left_col, self.time_block.right_col - 1):
                if date_utilities.parse_row_as_date(list(self.parent_sheet.raw_values.columns())[i][self.time_block.upper_row:self.time_block.lower_row]).equals(
                        date_utilities.parse_row_as_date(list(self.parent_sheet.raw_values.columns())[i + 1][self.time_block.upper_row:self.time_block.lower_row])):
                    current_pattern += 1
                else:
                    offsets.append(current_pattern)
                    current_pattern = 0

        else:
            current_pattern = 0
            for i in range(self.time_block.upper_row, self.time_block.lower_row - 1):
                if date_utilities.parse_row_as_date(self.parent_sheet.raw_values[i][self.time_block.left_col:self.time_block.right_col]).equals(date_utilities.parse_row_as_date(self.parent_sheet.raw_values[i + 1][self.time_block.left_col:self.time_block.right_col])):
                    current_pattern += 1
                else:
                    offsets.append(current_pattern)
                    current_pattern = 0
        # check for regularity of the given offset
        for i in range(len(offsets)-1):
            if offsets[i] != offsets[i+1]:
                return False, 0
        # Case when offset list is empty. Happens when all the dates are same in a range. Eg. [2007, 2007, 2007]
        if len(offsets) == 0:
            return False, 0

        return True, offsets[0]

    # check for the effect of merged time cells. non zero offset means multiple tables. Now only regular case is supported
    def check_offset(self):
        if self.time_orientation == utility.row_orientation:

            if self.has_equivalent_time_columns():
                regular, offset = self.has_regular_offset()
                if regular:
                    self.offset = offset
        else:
            if self.has_equivalent_time_rows():
                regular, offset = self.has_regular_offset()
                if regular:
                    self.offset = offset

    def find_granularity(self):
        if self.time_orientation == utility.column_orientation:
            self.granularity = date_utilities.find_granularity(list(self.parent_sheet.raw_values.rows())[self.time_block.upper_row:self.time_block.lower_row],
                                                          self.time_block.left_col, self.time_block.right_col)
        else:
            self.granularity = date_utilities.find_granularity(
                list(self.parent_sheet.raw_values.columns())[self.time_block.left_col:self.time_block.right_col],
                self.time_block.upper_row, self.time_block.lower_row)
            logging.info("Granularity of %s is %s", str(self.time_block), self.granularity)

    # find the granularity of the specified interval of the time block
    def find_bounded_granularity(self, start, end):
        if self.time_orientation == utility.column_orientation:
            self.granularity = date_utilities.find_granularity(list(self.parent_sheet.raw_values.rows())[start:end],
                                                          self.time_block.left_col, self.time_block.right_col)
        else:
            self.granularity = date_utilities.find_granularity(list(self.parent_sheet.raw_values.columns())[start:end],self.time_block.upper_row, self.time_block.lower_row)

        logging.info("Granularity of %s is %s", str(self.time_block), self.granularity)
        return self.granularity

    def get_orientation(self):
        return self.time_orientation

    def get_granularity(self):
        return self.granularity

    def add_label(self, label_block):
        self.labels.append(label_block)
        self.update_data_block(label_block)

    # detecting the number/empty blocks which are data of the time serieses
    def find_series_block(self):
        start, end = block_detector.find_data_block(self.parent_sheet.raw_values, self.parent_sheet.get_tags(), self.time_block.upper_row, self.time_block.lower_row, self.time_block.left_col,
                                                              self.time_block.right_col, self.time_orientation)
        self.data_start = start
        self.data_end = end

    # if there are some labels found and data was not found for them then should update the location of time series data
    def update_data_block(self, label_block):
        if self.time_orientation == utility.row_orientation:
            if label_block.lower_row > self.data_end:
                self.data_end = label_block.lower_row
        else:
            if label_block.right_col > self.data_end:
                self.data_end = label_block.right_col

    # fill the empty cells in the time block.[the back_fill mode case] (should be based on orientation)
    def fill_time_block(self):
        for i in range(self.time_block.upper_row, self.time_block.lower_row):
            for j in range(self.time_block.left_col, self.time_block.right_col):
                if self.parent_sheet.get_tags()[i][j] == {utility.empty_cell}:
                    # go and find the closest full cell and get the same value for that
                    row, col = block_detector.find_closest_date_cell(self.parent_sheet.get_tags(), i, j, self.time_block)
                    self.parent_sheet.raw_values[i, j] =  self.parent_sheet.raw_values[row, col]
                    self.parent_sheet.classified_tags[i][j] = self.parent_sheet.classified_tags[row][col]
                    self.back_mode_time = True

    def find_orientation(self):
        self.fill_time_block()
        if self.time_block.right_col - self.time_block.left_col == 1:
            return utility.column_orientation

        if self.time_block.lower_row - self.time_block.upper_row == 1:
            return utility.row_orientation
        else:
            logging.info("finding orientation in the worst case")
            # if row is a meaningful absolute time then it can be in columns
            if date_utilities.valid_time_row(self.parent_sheet.raw_values[self.time_block.upper_row][self.time_block.left_col:self.time_block.right_col])[0]:
                logging.error("time row was not consistent!")
                return utility.column_orientation

        return utility.row_orientation

    def find_table_borders(self, tags):
        table_borders = block_detector.rectangular_block()
        table_borders.left_col = self.time_block.left_col
        table_borders.right_col = self.time_block.right_col
        table_borders.upper_row = self.time_block.upper_row
        table_borders.lower_row = self.time_block.lower_row

        for l_block in self.labels:
            table_borders.left_col = min(table_borders.left_col , l_block.left_col)
            table_borders.right_col = max(table_borders.right_col, l_block.right_col)
            table_borders.upper_row = min(table_borders.upper_row, l_block.upper_row)
            table_borders.lower_row = max(table_borders.lower_row, l_block.lower_row)
        if self.time_orientation == utility.column_orientation:
            table_borders.left_col = min(table_borders.left_col, self.data_start)
            table_borders.right_col = max(table_borders.right_col, self.data_end)
        else:
            table_borders.upper_row = min(table_borders.upper_row, self.data_start)
            table_borders.lower_row = max(table_borders.lower_row, self.data_end)
        # continue from side untill reaching a place where it is totaly empty
        while table_borders.upper_row >= 0:
            if block_detector.is_empty_row(tags[table_borders.upper_row][table_borders.left_col:table_borders.right_col]):
                break
            table_borders.upper_row -= 1

        while table_borders.lower_row < len(tags):
            if block_detector.is_empty_row(tags[table_borders.lower_row][table_borders.left_col:table_borders.right_col]):
                break
            table_borders.lower_row += 1

        while table_borders.left_col >= 0:
            if block_detector.is_empty_col(tags, table_borders.upper_row, table_borders.lower_row, table_borders.left_col):
                break
            table_borders.left_col -= 1

        while True:
            if block_detector.is_empty_col(tags, table_borders.upper_row, table_borders.lower_row, table_borders.right_col):
                break
            table_borders.right_col += 1

        table_borders.upper_row += 1
        table_borders.left_col += 1
        self.borders = table_borders

    # Find label header if it is present in table
    def find_label_names(self, tags):
        if self.time_orientation == utility.row_orientation:
            label_row_num = self.time_block.upper_row
            unknown_label_counter = 0
            for x in self.labels:
                for i in range(x.left_col, x.right_col):
                    if utility.text_cell in tags[label_row_num][i]:
                        label_name = self.parent_sheet.raw_values[label_row_num, i]
                        self.label_names[i] = "_".join(utility.data_to_string(label_name).split(" ")).lower()
                    else:
                        #TODO: Search for label
                        suffix = ""
                        if unknown_label_counter != 0:
                            suffix = "_" + str(unknown_label_counter)
                        self.label_names[i] = "label" + suffix
                        unknown_label_counter += 1
                    self.label_backfill_bool[i] = self.is_merged_label(i, x.upper_row, x.lower_row)
                        
        else:
            label_counter = 0
            for x in self.labels:
                for i in range(x.upper_row, x.lower_row):
                    suffix = ""
                    if label_counter != 0:
                        suffix = "_" + str(label_counter)

                    self.label_names[i] = "label" + suffix
                    label_counter += 1

                    self.label_backfill_bool[i] = self.is_merged_label(i, x.left_col, x.right_col)

    #Check if label block has multiple cells merged together
    def is_merged_label(self, row, col_start, col_end):
        if self.time_orientation == utility.row_orientation:
            for i in range(col_start, col_end):
                if self.parent_sheet.get_merged_block(i, row):
                    return True
            return False
        else:
            for i in range(col_start, col_end):
                if self.parent_sheet.get_merged_block(row, i):
                    return True
            return False

    def create_json(self):
        time_serieses = []
        logging.info("BREAK POINTS: " + str(self.break_points))

        for point in range(len(self.break_points)-1):
            # Stores the set of labels which are already written to the json.
            # Used here to avoid repetition
            seen_labels = {}

            time_series_region_json = dict()
            time_series_region_json['orientation'] = self.time_orientation
            time_series_region_json['metadata'] = []

            logging.info("TIME BLOCK: " + str(self.time_block))

            if self.time_orientation == utility.row_orientation:
                if self.offset == 0:
                    time_series_region_json['locs'] = parsed_table.str_representation(self.time_block.left_col,
                                                                              self.time_block.right_col,
                                                                            utility.column_orientation)
                else:
                    time_series_region_json['locs'] = parsed_table.str_representation_offset(self.time_block.left_col,
                                                                                      self.time_block.right_col,
                                                                                      utility.column_orientation, self.offset+1)
                for x in self.labels:
                    logging.info("LABEL: " + str(x))
                    for i in range(x.left_col, x.right_col):
                        if i in seen_labels:
                            continue
                        else:
                            seen_labels[i] = True

                        meta = {"source": utility.column_orientation, "name": self.label_names[i],"loc": parsed_table.str_representation(i, i + 1, utility.column_orientation)}
                        if self.label_backfill_bool[i] == True:
                            meta["mode"] = "backfill"
                        time_series_region_json['metadata'].append(meta)

                time_series_region_json['times'] = {
                    "locs": parsed_table.str_representation(self.time_block.upper_row, self.time_block.lower_row,utility.row_orientation),"granularity": self.granularity}

                logging.error("time series region" + str(self.data_start) + "' " + str(self.data_end))
            else:
                if self.offset == 0:
                    time_series_region_json['locs'] = parsed_table.str_representation(self.break_points[point],self.break_points[point+1],utility.row_orientation)
                else:
                    time_series_region_json['locs'] = parsed_table.str_representation_offset(self.break_points[point],
                                                                                      self.break_points[point + 1],
                                                                                      utility.row_orientation, self.offset+1)

                for x in self.labels:
                    logging.info("LABEL: " + str(x))
                    for i in range(x.upper_row, x.lower_row):
                        if i in seen_labels:
                            continue
                        else:
                            seen_labels[i] = True

                        meta = {"source": utility.row_orientation, "name": self.label_names[i],"loc": parsed_table.str_representation(i, i + 1, utility.row_orientation)}
                        if self.label_backfill_bool[i] == True:
                            meta["mode"] = "backfill"
                        time_series_region_json['metadata'].append(meta)
                # if self.offset == 0:
                time_series_region_json['times'] = {
                "locs": parsed_table.str_representation(self.time_block.left_col, self.time_block.right_col,utility.column_orientation),
                "granularity": self.find_bounded_granularity(self.break_points[point], self.break_points[point+1])}
                # else:
                #     time_series_region_json['times'] = {
                #         "locs": parsed_table.str_representation_offset(self.time_block.left_col, self.time_block.right_col,
                #                                                 utility.column_orientation, self.offset+1),
                #         "granularity": self.find_bounded_granularity(self.break_points[point],
                #                                                      self.break_points[point + 1])}
            if self.back_mode_time == True:
                time_series_region_json['times']['back_fill_mode'] = True
            logging.error("time series region" + str(self.data_start) + "' " + str(self.data_end))
            time_series_region_json[self.time_orientation + 's'] = parsed_table.str_representation(self.data_start,self.data_end,self.time_orientation)

            if len(time_series_region_json['metadata']) >= 15:
                logging.error("Time series region: ", time_series_region_json[self.time_orientation + 's'], " ", time_series_region_json['locs'], " is likely not valid.")
                continue
            time_serieses.append(time_series_region_json)

        return time_serieses

    def find_time_block_length(self):
        time_block = self.time_block
        time_header_length = max(time_block.lower_row - time_block.upper_row, time_block.right_col - time_block.left_col)
        return time_header_length

    # string representation of the given interval
    @classmethod
    def str_representation(cls, start, end, orientation):
        interval = "["
        if orientation == utility.row_orientation:
            if start+1 == end:
                interval += str(start+1) + "]"
            else:
                interval = interval + str(start+1) + ":" + str(end) + "]"
        else:
            if start+1 == end:
                interval += parsed_table.get_excel_column_name(start) + "]"
            else:
                interval = interval + parsed_table.get_excel_column_name(start) + ":" + parsed_table.get_excel_column_name(end-1) + "]"

        return interval
    @classmethod
    def str_representation_offset(cls, start, end, orientation, offset):
        interval = "["
        if orientation == utility.row_orientation:
            if start+1 == end:
                interval += str(start+1) + "]"
            else:
                interval = interval + str(start+1) + ":" +str(offset)+":"+ str(end) + "]"
        else:
            if start+1 == end:
                interval += parsed_table.get_excel_column_name(start) + "]"
            else:
                interval = interval + parsed_table.get_excel_column_name(start) + ":" +str(offset)+":"+ parsed_table.get_excel_column_name(end-1) + "]"
        return interval

    @classmethod
    def get_excel_column_name(cls, idx):
        if idx <= 25:
            return chr(ord('A') + idx)
        return parsed_table.get_excel_column_name(idx/26 - 1) + chr(ord('A') + idx%26)

    @classmethod
    def is_valid_time_header(self, time_block):
        if time_block.upper_row + 1 == time_block.lower_row and time_block.left_col + 1 == time_block.right_col:
            return False
        if time_block.upper_row + 1 > time_block.lower_row or time_block.left_col + 1 > time_block.right_col:
            logging.warn("Invalid time header skipping")
            return False
        return True

    #Check if two blocks overlap each other
    def is_overlapping(self, table_borders, date_block):
        if date_block.left_col >= table_borders.right_col:
            return False
        if date_block.right_col <= table_borders.left_col:
            return False
        if date_block.upper_row >= table_borders.lower_row:
            return False
        if date_block.lower_row <= table_borders.upper_row:
            return False

        return True

    def date_alignment_measure(self, table):
        if self.time_orientation != table.get_orientation():
            return 0

        old_start = old_end = new_start = new_end = -1
        if self.time_orientation == utility.row_orientation:
            old_start = table.time_block.left_col
            old_end   = table.time_block.right_col
            new_start = self.time_block.left_col
            new_end   = self.time_block.right_col
        else:
            old_start = table.time_block.upper_row
            old_end   = table.time_block.lower_row
            new_start = self.time_block.upper_row
            new_end   = self.time_block.lower_row

        # Computing score similar to F1 score for alignment
        # KNOWN TIME BLOCK:    |-----------|-----------------|                |
        # UNKNOWN TIME BLOCK:  |     FN    |-------TP--------|-------FP-------|
        # FN - False Negative
        # TP - True Positive
        # FP - False Positive

        TP = min(old_end, new_end) - max(old_start, new_start)
        FP = max(0, new_end - old_end) + max(0, old_start - new_start)
        FN = max(0, old_end - new_end) + max(0, new_start - old_start)

        P = float(TP)/(TP + FP)
        R = float(TP)/(TP + FN)

        if P + R != 0:
            score = 2 * P * R / (P + R)
        else:
            score = 0

        return score

    #Validate time block
    def is_valid_time_header_post(self, known_tables):
        #Heuristics to check whether series of numbers is a date
        #H1: Date block has a different orientation from previous date blocks
        #H2: Date block is shorter compared to previous blocks
        
        #H3: Check surroundings of date blocks, position in table
        #Case 1: Date block is outside of previously detected tables => probably a date
        #Case 2: dates are aligned but inside previous table => multiple tables are merged to form a single table structure
        #Case 3: dates are not aligned and inside previous table => It's not actually a date.
        
        different_orientation = True
        short_time_block = True
        table_overlap = False
        max_date_alignment_measure = 0

        if len(known_tables) == 0:
            different_orientation = False
            short_time_block = False
        
        for table in known_tables:
            #H1
            if self.time_orientation == table.get_orientation():
                different_orientation = False
            #H3
            if self.is_overlapping(table.borders, self.time_block):
                table_overlap = True
            date_alignment_measure = self.date_alignment_measure(table)
            max_date_alignment_measure = max(max_date_alignment_measure, date_alignment_measure)

        #H2
        if self.time_block_length >= 5:
            short_time_block = False
        else:
            for table in known_tables:
                if table.time_block_length < 5:
                    short_time_block = False

        if table_overlap == False:
            return True
        elif date_alignment_measure >= 0.666:
            logging.info("Previous table border may have been computed incorrectly")
            return True
        elif different_orientation == True:
            return False
        elif short_time_block == True:
            logging.info("Highly unlikely to be a time block, but not sure")
            return False
        else:
            logging.info("Unexpected date block found.")
            return True



import unittest
from etk.wikidata.utils import *


class TestWikidataUtils(unittest.TestCase):
    def test_parse_datetime_string(self):
        self.assertEqual(
            parse_datetime_string('19710101 01:02:03', additional_formats=['%Y%m%d %H:%M:%S']),
            ('1971-01-01T01:02:03', Precision.second))
        self.assertEqual(
            parse_datetime_string('19710101 01:02', additional_formats=['%Y%m%d %H:%M']),
            ('1971-01-01T01:02:00', Precision.minute))
        self.assertEqual(
            parse_datetime_string('19710101 1am', additional_formats=['%Y%m%d %Ham']),
            ('1971-01-01T01:00:00', Precision.hour))
        self.assertEqual(
            parse_datetime_string('19710101'),
            ('1971-01-01T00:00:00', Precision.day))
        self.assertEqual(
            parse_datetime_string('197101'),
            ('1971-01-01T00:00:00', Precision.month))
        self.assertEqual(
            parse_datetime_string('1971', additional_formats=['%Y']),
            ('1971-01-01T00:00:00', Precision.year))



import json
import logging
import argparse

logging.basicConfig(format='%(name)s - %(levelname)s - %(message)s', level=logging.INFO)


def process(filename):

    out_file = open("output.txt", mode='w')
    step = 1
    final_json = []

    for sheet, sheet_name, merged_cells in input_processor.process_excel(filename):
        logging.info("Processing sheet: %s", sheet_name)

        ps = sheet_processor.parsed_sheet(sheet_name, step, sheet, merged_cells)
        step += 1
        ps.find_tables(out_file)
        tmp = ps.get_output()
        if tmp == None:
            continue
        final_json.append(tmp)
        logging.debug(json.dumps(ps.get_output(), sort_keys=True, indent=2, separators=(',', ': ')))

    return final_json


def main():
    ap = argparse.ArgumentParser()

    ap.add_argument("infile", help='File to annotate.')
    ap.add_argument("--outfile", help='File to write the annotation to.', default='')
    args = ap.parse_args()

    infile = args.infile.replace("\\", "")
    outfile = args.outfile.replace("\\", "")  # Handle case when outfile is empty

    if outfile == "":
        outfile = infile + "_annotation.json"

    final_json = process(infile)

    with open(outfile, mode='w') as out:
        json.dump(final_json, out, sort_keys=True, indent=2, separators=(',', ': '))
        logging.info("Output written to " + outfile)


if __name__ == "__main__":
    main()



import unittest
from rdflib import Graph, URIRef
import rdflib.namespace as rlns
from etk.knowledge_graph.namespacemanager import NamespaceManager, SCHEMA, DIG
from etk.knowledge_graph.namespacemanager import WrongFormatURIException, PrefixNotFoundException
from etk.knowledge_graph.node import URI


class TestKGNamespaceManager(unittest.TestCase):
    def test_namespace_parse_uri(self):
        nm = NamespaceManager(Graph())
        nm.bind('dig', DIG)
        uri = URIRef('http://dig.isi.edu/ontologies/dig/Event')
        self.assertEqual(uri, nm.parse_uri(uri))
        for case in ('http://dig.isi.edu/ontologies/dig/Event', 'dig:Event'):
            self.assertEqual(uri, nm.parse_uri(case))
        with self.assertRaises(PrefixNotFoundException):
            nm.parse_uri(':Event')
        nm.bind('', 'http://dig.isi.edu/ontologies/dig/')
        self.assertEqual(uri, nm.parse_uri(':Event'))
        self.assertEqual(uri, nm.parse_uri('Event'))
        self.assertEqual(uri, nm.parse_uri(URI(':Event')))
        with self.assertRaises(WrongFormatURIException):
            nm.parse_uri(None)

    def test_namespace_bind(self):
        correct_content = '@prefix : <http://w.org/> . @prefix schema: <http://schema.org/> .'
        replace_content = '@prefix schema: <http://dig.schema.org/> .'
        nm = NamespaceManager(Graph())
        nm.bind('owl', rlns.OWL)
        nm.bind('schema', SCHEMA)
        nm.bind('dig', DIG)
        namespace = {x[0] for x in nm.namespaces()}
        self.assertNotIn('', namespace)
        for name in ('owl', 'rdfs', 'rdf', 'schema', 'xsd', 'xml', 'dig'):
            self.assertIn(name, namespace)

        nm.graph.parse(data=correct_content, format='ttl')
        namespace = {x[0] for x in nm.namespaces()}
        self.assertIn('', namespace)

        nm.graph.parse(data=replace_content, format='ttl')
        namespace = {x[0] for x in nm.namespaces()}
        self.assertIn('schema', namespace)
        self.assertNotEqual(nm.store.namespace('schema'), SCHEMA)
        self.assertEqual(nm.store.namespace('schema'), URIRef('http://dig.schema.org/'))

    def test_namespace_split_uri(self):
        nm = NamespaceManager(Graph())
        nm.bind('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        ns, property_ = nm.split_uri('http://www.w3.org/1999/02/22-rdf-syntax-ns#label')
        self.assertEqual(str(ns), 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        self.assertEqual(nm.store.prefix(ns), 'rdf')
        self.assertEqual(property_, 'label')

    def test_namespace_check_uriref(self):
        nm = NamespaceManager(Graph())
        uri = URIRef('http://dig.isi.edu/ontologies/dig/Event')
        self.assertEqual(nm.check_uriref(uri), uri)

        text = 'http://dig.isi.edu/ontologies/dig/Event'
        self.assertEqual(nm.check_uriref(text), uri)

        text = URI('http://dig.isi.edu/ontologies/dig/Event')
        self.assertEqual(nm.check_uriref(text), uri)



from etk.timeseries.annotation import utility, block_detector
from etk.timeseries.annotation.table_processor import parsed_table
from etk.timeseries.annotation import cell_classifier
import logging
from etk.timeseries.annotation import multi_table_processor

class parsed_sheet:
    def __init__(self, name, index, real_values, merged_cells):
        self.sheet_name = name
        self.sheet_index = index
        self.raw_values = real_values
        self.classifier = cell_classifier.simple_tag_classifier()
        self.add_tags()
        self.table_list = []
        self.merged_cells = merged_cells
        self.in_complex = False


    # Returns the merged block if the given cell is located in one.
    def get_merged_block(self, row, col):
        for block in self.merged_cells:
            if row >= block[0] and row < block[1] and col >= block[2] and col < block[3]:
                return block
        return None

    # looks for the neighbor cells of the given date_cell and returns the the smallest rectangular block that this date is located in.
    def find_date_blocks(self, date_cell):
        blocks = []
        for i in range(len(self.classified_tags)):
            for j in range(len(self.classified_tags[i])):
                if date_cell not in self.classified_tags[i][j]:
                    continue
                if block_detector.check_cell_in_block(blocks, i, j):
                    continue
                blocks.append(block_detector.BFS_date_block(i, j, self.classified_tags, date_cell))

        for block in blocks:
            logging.debug("Date block: " + str(block))
            yield block_detector.find_minimum_boundry_rectangle(block)

    # finding the tables with time series in the given sheet
    def find_tables(self, out_file):
        step = 1
        sheet_label_blocks = []
        logging.info("------------------------------------")
        for rb in self.find_date_blocks(self.classifier.get_date_tag()):
            # check if the header is only a single time cell (which can be an event in the best case)

            if not parsed_table.is_valid_time_header(rb):
                print("Pre validation of time header failed. " + str(rb))
                continue

            detected_table = parsed_table(rb, self)
            if not detected_table.is_valid_time_header_post(self.table_list):
                print("Post validation of time header failed. " + str(rb))
                continue

            sheet_label_blocks = []
            step += 1
            orientation = detected_table.get_orientation()

            print("Date block detected : " + str(rb))
            logging.info("Orientation is " + orientation)

            # if they are merged time cells
            for i in range(rb.upper_row, rb.lower_row):
                for j in range(rb.left_col, rb.right_col):
                    if self.get_merged_block(i, j) != None:
                        detected_table.check_offset()

            # check for overlapping tables caused by overlapping times
            has_overlap, break_points = multi_table_processor.complex_table.check_for_overlapping_time(detected_table.time_block, orientation, self.raw_values)
            detected_table.break_points = break_points
            if has_overlap:
                detected_table.is_complex = True

            if orientation == utility.row_orientation:
                for x in block_detector.find_label_block(self.get_tags(), detected_table.data_start, rb.lower_row, 0, rb.left_col, utility.column_key, sheet_label_blocks):
                    detected_table.add_label(x)
            else:

                for x in block_detector.find_label_block(self.get_tags(), detected_table.data_start, rb.right_col, 0, rb.upper_row, utility.row_key, sheet_label_blocks):
                    out_file.write('rows[' + str(x.upper_row + 1) + "-" + str(x.lower_row) + "] columns[ " + str( x.left_col + 1) + "-" + str(x.right_col) + "]\n")

                    detected_table.add_label(x)

            detected_table.find_table_borders(self.get_tags())
            logging.info("Table borders: " + str(detected_table.borders))

            # check for mergeds cells aftar finding the table borders
            # check for the remaining cells that are not present in any block. They may be labels?
            # check for unassigned_cells

            dtb = detected_table.borders
            #Find more labels on the right/bottom side of date blocks
            if orientation == utility.row_orientation:
                for x in block_detector.find_label_block(self.get_tags(), detected_table.data_start, rb.lower_row, rb.right_col, dtb.right_col, utility.column_key, sheet_label_blocks):
                    detected_table.add_label(x)
                    out_file.write('rows[' + str(x.upper_row + 1) + "-" + str(x.lower_row) + "] columns[ " + str(x.left_col + 1) + "-" + str(x.right_col) + "]\n")
            else:
                for x in block_detector.find_label_block(self.get_tags(), detected_table.data_start, rb.right_col, rb.lower_row, dtb.lower_row, utility.row_key, sheet_label_blocks):
                    out_file.write('rows[' + str(x.upper_row + 1) + "-" + str(x.lower_row) + "] columns[ " + str( x.left_col + 1) + "-" + str(x.right_col) + "]\n")
                    detected_table.add_label(x)
            detected_table.find_label_names(self.get_tags())

            self.add_table(detected_table)

    def add_table(self, table):
        self.table_list.append(table)


    def add_tags(self):
        self.classified_tags = self.classifier.tag_sheet(self.raw_values)


    # get class tags of the sheet cells
    def get_tags(self):
        return self.classified_tags

    # creates the output json for the sheet based on the tables found
    def get_output(self):
        if len(self.table_list) == 0:
            return None
        sheet_json = dict()
        property = {"sheet_indices": str([self.sheet_index])}
        sheet_json['Properties'] = property
        sheet_json['TimeSeriesRegions'] = []
        for table in self.table_list:
            for t_json in table.create_json():
                sheet_json['TimeSeriesRegions'].append(t_json)
        sheet_json['GlobalMetadata'] = [{"name": "title", "source": "sheet_name"}]

        return sheet_json

    @classmethod
    def merge_sheet_indices(cls, idx1, idx2):
        idx1_stripped = idx1[1:-1].split(":")
        idx2_stripped = str(idx2[1:-1])

        if idx1 == idx2:
            return idx1
        elif len(idx1_stripped) == 2 and int(idx1_stripped[1]) + 1 == int(idx2_stripped):
            return "[" + idx1_stripped[0] + ":" + idx2_stripped + "]"
        elif len(idx1_stripped) == 1 and int(idx1_stripped[0]) + 1 == int(idx2_stripped):
            return "[" + idx1_stripped[0] + ":" + idx2_stripped + "]"
        else:
            raise Exception("Sheet indices cannot be merged.")

    @classmethod
    def merge_row_indices(cls, idx1, idx2):
        if idx1 == idx2:
            return idx1
        else:
            raise Exception("Row indices cannot be merged.")

    @classmethod
    #Merge json for 2 sheets if they are similar
    def merge_json(cls, json1, json2):
        try:
            if type(json1) != type(json2):
                raise Exception("Json type is not equal")
            if isinstance(json1, list):
                ret = []
                if len(json1) != len(json2):
                    raise Exception("List length not equal")
                for item1, item2 in zip(json1, json2):
                    merged = cls.merge_json(item1, item2)
                    ret.append(merged)
                return ret
            elif isinstance(json1, dict):
                ret = {}
                if len(json1) != len(json2):
                    raise Exception("Dict length not equal.")
                for key in json1:
                    if key in json2:
                        if type(json1[key]) == type(json2[key]):
                            if isinstance(json1[key], list) or isinstance(json1[key], dict):
                                merged = cls.merge_json(json1[key], json2[key])
                                ret[key] = merged
                            elif key == "sheet_indices":
                                ret[key] = cls.merge_sheet_indices(json1[key], json2[key])
                            elif key == "rows":
                                ret[key] = cls.merge_row_indices(json1[key], json2[key])
                            elif json1[key] == json2[key]:
                                ret[key] = json1[key]
                            else:
                                raise Exception("Key values are not equal.")
                        else:
                            raise Exception("Dict key types do not match")
                    else:
                        raise Exception("Key not found in dict")
                return ret
            elif json1 == json2:
                return json1
        
        except Exception as exc:
            raise Exception(exc)

    @classmethod
    def merge_json_list(cls, final_json):
        return_json = []
        json = final_json[0]
        for i in range(len(final_json)):
            try:
                merged_json = cls.merge_json(json, final_json[i])
                json = merged_json
            except:
                return_json.append(json)
                json = final_json[i]

        return_json.append(json)

        return return_json



from etk.wikidata.truthy import TruthyUpdater


def add_arguments(parser):
    parser.add_argument('-e', '--endpoint', action='store', type=str, dest='endpoint')
    parser.add_argument('-f', '--input', action='store', type=str, dest='file')
    parser.add_argument('--user', action='store', default=None, type=str, dest='user')
    parser.add_argument('--passwd', action='store', default=None, type=str, dest='passwd')
    parser.add_argument('--dry-run', action='store_true', default=False, dest='dryrun')


def run(args):
    print('Start updating endpoint: {}'.format(args.endpoint))
    if args.dryrun:
        print('Mode: Dryrun')
    tu = TruthyUpdater(args.endpoint, args.dryrun, args.user, args.passwd)
    np_list = []
    with open(args.file) as f:
        for l in f.readlines():
            if not l: continue
            node, prop = l.strip().split('\t')
            np_list.append((node, prop))
    tu.build_truthy(np_list)
    print('Update finished!')



from etk.knowledge_graph.graph import Graph
from etk.knowledge_graph.knowledge_graph import KnowledgeGraph
from etk.knowledge_graph.namespacemanager import NamespaceManager
from etk.knowledge_graph.node import Node, URI, BNode, Literal, LiteralType
from etk.knowledge_graph.ontology import Ontology
from etk.knowledge_graph.schema import KGSchema
from etk.knowledge_graph.subject import Subject, Reification



import unittest
import pyshacl
import rdflib
from etk.knowledge_graph.shacl import SH, SHACL, SHACLOntoConverter
from etk.knowledge_graph.ontology import Ontology
from etk.knowledge_graph.node import URI, BNode, Literal, LiteralType
from etk.knowledge_graph.subject import Subject
from etk.knowledge_graph.graph import Graph


def subject_to_graph(subject):
    g = Graph()
    g.bind('sh', SH)
    g.bind(None, 'http://example.org/')
    g.add_subject(subject)
    return g


namespaces = {
    '': 'http://example.org/',
    'sh': SH,
    'xsd': rdflib.XSD
}


class TestPySHACL(unittest.TestCase):
    def setUp(self):
        content = """
        @prefix : <http://isi.edu/xij-rule-set#> .
        :etk a :Software ;
            :developer [ a :Developer ;
                    :name "Sylvia", "Qingyuandi" ],
                [ a :Developer ;
                    :name "Amandeep" ],
                [ a :Developer ;
                    :name "Dongyu" ],
                [ a :Developer ;
                    :name "Runqi" ] .

        :rltk a :Software ;
            :developer [ a :Developer ;
                    :name "Pedro" ],
                [ a :Developer ;
                    :name "Yixiang" ],
                [ a :Developer ;
                    :name "Mayank" ] .
        """
        self.data_graph = rdflib.Graph()
        self.data_graph.parse(data=content, format='ttl')

        ontology = """
            @prefix : <http://isi.edu/xij-rule-set#> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :Person a owl:Class .
            :Developer a owl:Class ;
                       rdfs:subClassOf :Person .
            :name a owl:DatatypeProperty ;
                  rdfs:domain :Person ;
                  rdfs:range xsd:string .
        """
        self.onto_graph = rdflib.Graph()
        self.onto_graph.parse(data=ontology, format='ttl')

        shacl_content = """
        @prefix : <http://example.org/> .
        @prefix xij: <http://isi.edu/xij-rule-set#> .
        @prefix sh: <http://www.w3.org/ns/shacl#> .
        @prefix xsd: <http://www.w3.org/2001/XMLSchema#> . 
        :SoftwareShape a sh:NodeShape ;
                       sh:targetClass xij:Software ;
                       sh:property [ 
                         sh:path xij:developer ;
                         sh:minCount 1 ;
                         sh:class xij:Developer ;
                       ] .
        :PersonShape a sh:NodeShape ;
                        sh:targetClass xij:Person ;
                        sh:property [
                          sh:path xij:name ;
                          sh:minCount 1 ;
                          sh:maxCount 1 ;
                          sh:datatype xsd:string ;
                        ] .
        """
        self.shacl_graph = rdflib.Graph()
        self.shacl_graph.parse(data=shacl_content, format='ttl')

    def test_pyshacl(self):
        conforms, results_graph, results_text = pyshacl.validate(self.data_graph, shacl_graph=self.shacl_graph,
                                                                 inference='rdfs', abort_on_error=False,
                                                                 meta_shacl=False, debug=False)
        self.assertTrue(conforms)

    def test_pyshacl_with_ontology(self):
        conforms, results_graph, results_text = pyshacl.validate(self.data_graph + self.onto_graph,
                                                                 shacl_graph=self.shacl_graph, inference='rdfs',
                                                                 abort_on_error=False, meta_shacl=False, debug=False)
        self.assertFalse(conforms)


class TestSHACLOntoConverter(unittest.TestCase):
    def test_cnode(self):
        converter = SHACLOntoConverter()
        c1 = rdflib.URIRef('http://example.org/Person')
        c2 = rdflib.URIRef('http://example.org/Developer')
        cnode1 = converter.cnode(c1)
        cnode2 = converter.cnode('http://example.org/Person')
        cnode3 = converter.cnode(c2)
        self.assertIsInstance(cnode1, BNode)
        self.assertIsInstance(cnode2, BNode)
        self.assertIsInstance(cnode3, BNode)
        self.assertEqual(cnode1, cnode2)
        self.assertNotEqual(cnode1, cnode3)

        converter = SHACLOntoConverter({str(c1): URI(':Person')})
        cnode4 = converter.cnode(c1)
        self.assertEqual(cnode4, URI(':Person'))

    def test_add_list(self):
        converter = SHACLOntoConverter()
        list_subject = Subject(BNode())
        list_subject.add_property(URI(':list'), converter._add_list([URI(':a'), URI(':b'), URI(':c')]))
        list_graph = subject_to_graph(list_subject)
        self.assertTrue(list_graph._g.query("""
            PREFIX : <http://example.org/> 
            ASK { 
                ?s :list ( ?a ?b ?c ) ;
            }
        """))

    def test_property_shape(self):
        ontology = """
            @prefix : <http://example.org/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :name a owl:DatatypeProperty ;
              rdfs:range xsd:string ; .
            :developer a owl:ObjectProperty ;
              rdfs:range :PythonDeveloper, :JavaDeveloper ; .
        """
        onto_graph = Ontology()
        onto_graph.parse(ontology)
        converter = SHACLOntoConverter()
        converter.onto_graph = onto_graph._g
        p_shape = converter._property_shape('http://example.org/name')
        self.assertIsInstance(p_shape, Subject)
        p_shape_graph = subject_to_graph(p_shape)
        self.assertTrue(p_shape_graph._g.query("""
            ASK { ?s sh:path :name ;
                     sh:datatype xsd:string .
            }
        """, initNs=namespaces))

        p_shape = converter._property_shape('http://example.org/developer')
        self.assertIsInstance(p_shape, Subject)
        p_shape_graph = subject_to_graph(p_shape)
        self.assertTrue(p_shape_graph._g.query("""
            ASK { 
                ?s sh:path :developer ;
                   sh:or ( [ sh:class ?c1 ] [ sh:class ?c2 ]) .
                FILTER (?c1 != ?c2)
            }
        """, initNs=namespaces))

    def test_convert_non_referenced_property(self):
        ontology = """
            @prefix : <http://example.org/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :name a owl:DatatypeProperty ;
              rdfs:domain :Person, :Software ;
              rdfs:range xsd:string ; .
            :developer a owl:ObjectProperty ;
              rdfs:range :PythonDeveloper, :JavaDeveloper ; .
        """
        onto_graph = Ontology()
        onto_graph.parse(ontology)
        converter = SHACLOntoConverter()
        converter.onto_graph = onto_graph._g
        converter._convert_ontology_non_referenced_property()
        self.assertTrue(converter._g._g.query("""
            ASK { ?s a sh:PropertyShape ;
                     sh:path :developer ;
                     sh:or ?range .
            }
        """, initNs=namespaces))
        self.assertFalse(converter._g._g.query("ASK { ?s sh:path :name }", initNs=namespaces))

    def test_convert_ontology_class_node_shape(self):
        ontology = """
            @prefix : <http://example.org/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :Developer rdfs:subClassOf :Person .
            :name a owl:DatatypeProperty ;
              rdfs:domain :Person, :Software ;
              rdfs:range xsd:string ; .
            :developer a owl:ObjectProperty ;
              rdfs:domain :Software ;
              rdfs:range :PythonDeveloper, :JavaDeveloper ; .
        """
        onto_graph = Ontology()
        onto_graph.parse(ontology)
        converter = SHACLOntoConverter()
        converter.onto_graph = onto_graph._g
        converter._convert_ontology_class_node_shape()
        self.assertTrue(converter._g._g.query("""
            ASK { ?s a sh:NodeShape ; sh:targetClass :Developer }
        """, initNs=namespaces))
        self.assertTrue(converter._g._g.query("""
            ASK { ?s sh:targetClass :Person ; sh:property ?p }
        """, initNs=namespaces))
        self.assertTrue(converter._g._g.query("""
            ASK { 
              ?s sh:targetClass :Software ; 
                 sh:property ?p1 ;
                 sh:property ?p2 
              FILTER (?p1 != ?p2)
            }
        """, initNs=namespaces))

    def test_convert_ontology_owl_restriction(self):
        ontology = """
            @prefix : <http://example.org/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :Developer rdfs:subClassOf [ a owl:Restriction ;
                owl:onProperty :name ;
                owl:allValuesFrom xsd:string ;
                owl:cardinality 1 
            ] ; .
            :Software rdfs:subClassOf [ a owl:Restriction ;
                    owl:onProperty :name ;
                    owl:allValuesFrom xsd:string ;
                    owl:cardinality 1
                ];
                rdfs:subClassOf [ a owl:Restriction ;
                    owl:onProperty :developer ;
                    owl:allValuesFrom :Developer ;
                    owl:minCardinality 1
                ]; .
        """
        onto_graph = Ontology()
        onto_graph.parse(ontology)
        converter = SHACLOntoConverter()
        converter.onto_graph = onto_graph._g
        converter._convert_ontology_owl_restriction()
        self.assertTrue(converter._g._g.query("""
            ASK { 
                ?s a sh:NodeShape ; 
                   sh:targetClass :Developer ; 
                   sh:property ?p .
                ?p sh:path :name ;
                   sh:count 1 ;
                   sh:datatype xsd:string .
            }
        """, initNs=namespaces))
        self.assertTrue(converter._g._g.query("""
            ASK { 
                ?s sh:targetClass :Software ; 
                   sh:property ?p .
                ?p sh:path :developer ;
                   sh:minCount 1 ;
                   sh:class :Developer .
            }
        """, initNs=namespaces))

    def test_convert_ontology(self):
        ontology = """
            @prefix : <http://example.org/> .
            @prefix owl: <http://www.w3.org/2002/07/owl#> .
            @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
            @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
            :Developer rdfs:subClassOf [ a owl:Restriction ;
                owl:onProperty :name ;
                owl:allValuesFrom xsd:string ;
                owl:cardinality 1 
            ] ; .
            :Software rdfs:subClassOf [ a owl:Restriction ;
                    owl:onProperty :name ;
                    owl:allValuesFrom xsd:string ;
                    owl:cardinality 1
                ];
                rdfs:subClassOf [ a owl:Restriction ;
                    owl:onProperty :developer ;
                    owl:allValuesFrom :Developer ;
                    owl:minCardinality 1
                ]; .
        """
        onto_graph = Ontology()
        onto_graph.parse(ontology)
        converter = SHACLOntoConverter()
        shacl_graph = converter.convert_ontology(onto_graph)
        self.assertIsInstance(shacl_graph, Graph)


class TestSHACL(unittest.TestCase):
    def test_graph_validation(self):
        pass



import unittest
from etk.wikidata.value import *
from etk.knowledge_graph import URI


class TestWikidataValue(unittest.TestCase):
    def test_entity_value(self):
        i = Item('Q34')
        self.assertEqual(i.value, URI('wd:Q34'))
        self.assertEqual(i.type, URI('wikibase:WikibaseItem'))
        self.assertIsNone(i.full_value)
        self.assertIsNone(i.normalized_value)

        p = Property('P69')
        self.assertEqual(p.value, URI('wd:P69'))
        self.assertEqual(p.type, URI('wikibase:WikibaseProperty'))
        self.assertIsNone(p.full_value)
        self.assertIsNone(p.normalized_value)

    def test_time_value(self):
        tv = TimeValue('2019-05-05T00:00:00', Item('Q1985727'), Precision.day, 0)
        self.assertEqual(tv.value, Literal('2019-05-05T00:00:00', type_=LiteralType.dateTime))
        self.assertIsInstance(tv.full_value, Subject)
        self.assertIsNone(tv.normalized_value)
        self.assertEqual(tv.type, URI('wikibase:Time'))

    def test_external_identifier(self):
        ei = ExternalIdentifier('/m/0282x', URLValue('http://g.co/kg/m/0282x'))
        self.assertEqual(ei.value, Literal('/m/0282x', type_=LiteralType.string))
        self.assertIsNone(ei.full_value)
        self.assertIsInstance(ei.normalized_value, URI)
        self.assertIsInstance(ei.normalized_value, URLValue)
        self.assertEqual(ei.type, URI('wikibase:ExternalId'))

    def test_quantity_value(self):
        qv = QuantityValue(1.5)
        print(qv._v_name())
        self.assertEqual(qv.value, Literal('1.5', type_=LiteralType.decimal))
        self.assertIsInstance(qv.full_value, Subject)
        self.assertIsInstance(qv.normalized_value, Subject)
        self.assertEqual(qv.normalized_value, qv.full_value)
        self.assertEqual(qv.type, URI('wikibase:Quantity'))

    def test_string_value(self):
        sv = StringValue('blabla')
        self.assertEqual(sv.value, Literal('blabla', type_=LiteralType.string))
        self.assertIsNone(sv.normalized_value)
        self.assertIsNone(sv.full_value)
        self.assertEqual(sv.type, URI('wikibase:String'))

    def test_url_value(self):
        uv = URLValue('http://www.isi.edu/')
        self.assertIsInstance(uv, URI)
        self.assertEqual(uv.value, URI('http://www.isi.edu/'))
        self.assertIsNone(uv.full_value)
        self.assertIsNone(uv.normalized_value)
        self.assertEqual(uv.type, URI('wikibase:Url'))

    def test_globe_coordinate(self):
        gc = GlobeCoordinate(45.0, -1, 1, Item('Q405'))
        self.assertIsInstance(gc.full_value, Subject)
        self.assertIsNone(gc.normalized_value)
        self.assertEqual(gc.type, URI('wikibase:GlobeCoordinate'))

    def test_monolingual_text(self):
        mt = MonolingualText('Gracias', lang='es')
        self.assertEqual(mt.value, Literal('Gracias', lang='es'))
        self.assertIsNone(mt.full_value)
        self.assertIsNone(mt.normalized_value)
        self.assertEqual(mt.type, URI('wikibase:Monolingualtext'))



from etk.timeseries.annotation import input_processor
import json
import numpy as np
import re
from etk.timeseries.annotation.table_processor import parsed_table
from etk.timeseries.annotation.granularity_detector import GranularityDetector


def is_number(value):
    if is_int(value) or is_float(value):
        return 1
    return 0


def is_empty(value):
    return (not value) or (value == ".")


def is_data(value):
    return is_number(value) or is_empty(value)


def is_int(value):
    if isinstance(value, int):
        return 1
    try:
        int(value)
        return 1
    except:
        return 0


def is_float(value):
    if isinstance(value, float):
        return 1
    try:
        float(value)
        return 1
    except:
        return 0


def is_string_literal(value):
    if isinstance(value, str):
        return True
    return False


def is_alpha(value):
    # Note the use of search instead of match
    # Match fails to match "1 kg"
    alpha_regex = re.compile(r'[a-z]', re.IGNORECASE)
    if is_string_literal(value) and alpha_regex.search(value):
        return 1
    return 0


def get_base_annotation(tp):
    annotation = """
        {
            "GlobalMetadata": [
                {
                    "name": "title",
                    "source": "sheet_name"
                }
            ],
            "Properties": {
                "sheet_indices": "[1]"
            },
            "TimeSeriesRegions": [
                {
                    "cols": "[%s]",
                    "locs": "[%s:%s]",
                    "metadata": [
                        %s
                    ],
                    "orientation": "col",
                    "times": {
                        "granularity": "%s",
                        "locs": "[%s]"
                    }
                }
            ]
        }
    """

    return json.loads(annotation % (tp['ts_column'], tp['start_row'], tp['end_row'],
                                    get_header_object(tp['header_present']),
                                    tp['granularity'],
                                    tp['time_column']))


def get_header_object(is_header_present=True):
    if is_header_present:
        return """
        {
            "loc": "[1]",
            "name": "label",
            "source": "row"
        }
        """
    else:
        return ""


def is_header_present(sheet):
    if len(sheet) == 0:
        return False

    if is_alpha(sheet[0][0]) and is_alpha(sheet[0][1]):
        return True


def remove_bom(sheet):
    try:
        if sheet[0][0].startswith("\uFEFF"):
            sheet[0][0] = sheet[0][0][1:]
    except:
        pass


class SimpleAnnotator:
    def __init__(self, infile):
        self.infile = infile

    def get_annotation_json(self):

        final_json = []

        for sheet, sheet_name, merged_cells in input_processor.process_excel(self.infile):
            table_properties = dict()

            sheet = sheet.to_array()
            sheet = np.array(sheet)

            remove_bom(sheet)

            row, col = sheet.shape

            # Two columns exactly
            assert col == 2

            # Check if header row is present
            header_present = is_header_present(sheet)
            start_row = 0
            if header_present:
                start_row = 1

            # Find the data block
            data_col = []
            for i in range(2):
                is_data_col = True
                for j in range(start_row, row):
                    if not is_data(sheet[j][i]):
                        is_data_col = False
                        break
                if is_data_col:
                    data_col.append(i)

            if len(data_col) == 2:
                logging.error("Ambiguous data. Multiple data columns present")
                break

            if len(data_col) == 0:
                logging.error("Cannot find data column")
                break

            data_col = data_col[0]
            # Other column is assumed to be date column
            date_col = data_col ^ 1

            table_properties['start_row'] = start_row + 1
            table_properties['end_row'] = row
            table_properties['time_column'] = parsed_table.get_excel_column_name(date_col)
            table_properties['header_present'] = header_present
            table_properties['ts_column'] = parsed_table.get_excel_column_name(data_col)
            table_properties['granularity'] = GranularityDetector.get_granularity(sheet[:, date_col])

            annotation = get_base_annotation(table_properties)

            final_json.append(annotation)

        return final_json



import unittest
from etk.timeseries.annotation.granularity_detector import GranularityDetector
from datetime import datetime, date


class TestGranularityDetector(unittest.TestCase):
    def test_granularity_detection(self):
        dates1 = ["2001-01-01", "2001-01-02", "2001-01-03", "2001-01-04"]
        dates2 = ["2001-01-01"]
        dates3 = ["2001", "2002", "2003", "2004", "2005"]
        dates4 = ["2001-01-01", "2001-04-01", "2001-07-01", "2001-10-01"]
        dates5 = ["2010-01-01", "2010-01-02", "2010-01-03", "cant parse this", "2010-01-05", "2010-01-06"]
        # dates6 = ["2001Q1", "2001Q2", "2001Q3"]

        assert GranularityDetector.get_granularity(dates1, return_best=True) == "day"
        assert GranularityDetector.get_granularity(dates2, return_best=True) == "unknown"
        assert GranularityDetector.get_granularity(dates3, return_best=True) == "year"
        assert GranularityDetector.get_granularity(dates4, return_best=True) == "quarter"
        assert GranularityDetector.get_granularity(dates5, return_best=True) == "day"
        # assert GranularityDetector.get_granularity(dates6, return_best=True) == "quarterly"

    def test_date_parser(self):
        date1 = "\ufeff2001-01-01"
        date2 = datetime(2001, 1, 1)
        date3 = None
        date4 = "garbage"
        date5 = 1
        date6 = date(2001, 1, 1)
        date7 = "2018-10-08 20:20:55.130874+00:00"

        assert GranularityDetector.get_parsed_date(date1) == datetime(2001, 1, 1)
        assert GranularityDetector.get_parsed_date(date2) == datetime(2001, 1, 1)
        assert GranularityDetector.get_parsed_date(date3) == None
        assert GranularityDetector.get_parsed_date(date4) == None
        assert GranularityDetector.get_parsed_date(date5) == None
        assert GranularityDetector.get_parsed_date(date6) == datetime(2001, 1, 1)
        assert GranularityDetector.get_parsed_date(date7) == datetime(2018, 10, 8, 20, 20,
                                                                      00)  # Seconds resolution is lost



from etk.timeseries.annotation.utility import date_cell, text_cell, empty_cell, blank_cell, number_cell, row_key, column_key, column_orientation
import logging

class rectangular_block:
    def __init__(self):
        self.upper_row = 0
        self.lower_row = 0
        self.left_col = 0
        self.right_col = 0

    def __str__(self):
        attrs = vars(self)
        return ', '.join("%s: %s" % item for item in attrs.items())

def find_label_block(tags, start_break, row_lim, col_start, col_end, orientation, detected_blocks):
    blocks = []
    tag_user = tags

    if orientation == row_key:
        #Transpose tag matrix
        tag_user = list(map(list, zip(*tags)))
        for rb in find_label_block(tag_user, start_break, row_lim, col_start, col_end, column_key, detected_blocks):
            t1, t2 = rb.upper_row, rb.lower_row
            rb.upper_row, rb.lower_row = rb.left_col, rb.right_col
            rb.left_col, rb.right_col = t1, t2
            yield rb
    else:
        curr_row = row_lim
        while curr_row < start_break or (curr_row < len(tags) and is_empty_label_row(tags[curr_row], col_end)):
            curr_row += 1
        # edit here to first go and look for the first non empty row that matches...
        while curr_row < len(tags) and not is_empty_label_row(tags[curr_row], col_end):
            curr_row += 1
            #Should we check till start of sheet?
            for col in range(col_end - 1, col_start - 1, -1): # should n't I only keep the closest?
                #Note: Removing this as labels are not necessarily text. Run against dataset and check TODO
                #if text_cell not in tags[curr_row-1][col]:
                #    continue
                if empty_cell in tags[curr_row - 1][col]:
                    continue
                if check_cell_in_block(detected_blocks, curr_row-1, col):
                    continue
                if check_cell_in_block(blocks, curr_row-1, col):
                    continue

                blocks.append(BFS_limited_block(curr_row-1, col, tags, row_lim, col_start, col_end))
    for block in blocks:
        minimum_boundary_rect = find_minimum_boundry_rectangle(block)
        logging.info("Label Block: " + str(minimum_boundary_rect))
        detected_blocks.append(block)
        yield minimum_boundary_rect


def is_empty_label_row(tags, col_lim):
    for i in range(col_lim):
        if empty_cell not in tags[i]:
            return False
        else: # or maybe analyze the meaning of the text in order to find out if it is the end or not
            continue
    return True


def BFS_limited_block(row, col, tags, row_lim, col_start, col_end):
    Q = []
    Q.append((row, col))
    block = []
    directions = [(0, 1), (0, -1), (1, 0), (-1, 0)]
    while len(Q) > 0:
        front = Q[0]
        block.append(Q.pop(0))
        # check the neighbors
        for dir in directions:
            if front[1] + dir[1] < len(tags[front[0]]) and front[1] + dir[1] < col_end and front[1] + dir[1] >= col_start and front[0] + dir[0] >= row_lim and front[0] + dir[0] < len(tags) and front[0] + dir[0] >= 0 and front[1] + dir[1] >= 0:
                #Removing check for text cell
                if empty_cell not in tags[front[0] + dir[0]][front[1] + dir[1]]:
                    if (front[0] + dir[0], front[1] + dir[1]) not in Q and (
                        front[0] + dir[0], front[1] + dir[1]) not in block:
                        Q.append((front[0] + dir[0], front[1] + dir[1]))

    return block


def find_data_block(a, tags, time_header_begin_row, time_header_end_row, time_header_begin_col, time_header_end_col, time_series_orientation):
    if time_series_orientation == column_orientation:
        #Transpose tag matrix
        tag_user = map(list, zip(*tags))
        return find_data_block(a, tag_user, time_header_begin_col, time_header_end_col, time_header_begin_row, time_header_end_row, row_key)
    else:
        # find the first non-empty row (should I add containing only numbers?)

        start_row = time_header_end_row
        for i in range(time_header_end_row, len(tags)): # continue untill reaching the first non empty row
            if is_a_time_series_row(tags[i], time_header_begin_col, time_header_end_col):
                start_row = i
                break

        j = start_row
        for j in range(start_row, len(tags)):
            if is_a_time_series_row(tags[j], time_header_begin_col, time_header_end_col):
                continue

            return start_row, j
        return start_row, j


def find_minimum_boundry_rectangle(block):
    first_cell = block[0]
    rb = rectangular_block()
    min_row = first_cell[0]
    min_col = first_cell[1]
    max_row = min_row
    max_col = min_col
    for cell in block:
        if cell[0] < min_row:
            min_row = cell[0]
        if cell[0] > max_row:
            max_row = cell[0]
        if cell[1] < min_col:
            min_col = cell[1]
        if cell[1] > max_col:
            max_col = cell[1]
    rb.left_col = min_col
    rb.right_col = max_col+1
    rb.upper_row = min_row
    rb.lower_row = max_row + 1
    return rb


def BFS_date_block(row, col, tags, desired_tag):
    logging.info("Starting BFS from [%d, %d]", row, col)
    Q = []
    Q.append((row, col))
    block = []
    directions = [(0,1), (0, -1), (1, 0), (-1, 0)]
    while len(Q) > 0:
        front = Q[0]
        block.append(Q.pop(0))
        # check the neighbors
        for dir in directions:
            if front[1] + dir[1] < len(tags[front[0]]) and front[0]+dir[0] < len(tags) and front[0]+dir[0] >= 0 and front[1] + dir[1] >= 0:
                if desired_tag in tags[front[0]+dir[0]][front[1]+dir[1]]:
                    if (front[0]+dir[0], front[1]+dir[1]) not in Q and (front[0]+dir[0], front[1]+dir[1]) not in block:
                        Q.append((front[0]+dir[0], front[1]+dir[1]))
    logging.info(block)
    return block

def check_in_merged_block(blocks, row, col):
    for block in blocks:
        if row >= block[0] and row < block[1] and col >= block[2] and col < block[3]:
            return True
    return False


def check_cell_in_block(blocks, row, col):
    for block in blocks:
        for cell in block:
            if cell[0] == row and cell[1] == col:
                return True
    return False


def is_a_time_series_row(row, left_col, right_col):
    seen_tags = row[left_col:right_col]
    seen_number = False

    # if number cell not available or cells are empty
    for x in seen_tags:
        if x == {text_cell} or x == {date_cell}:
            return False
        if number_cell in x:
            seen_number = True
            continue
        if blank_cell in x:
            seen_number = True
            continue
        if empty_cell in x:
            continue

    if not seen_number:
        return False
    return True

def is_empty_row(row_tags):
    for tag in row_tags:
        if empty_cell not in tag:
            return False
    return True


def is_empty_col(tags, start_row, end_row, col):
    for row in range(start_row, end_row):
        if col >= len(tags[row]):
            continue
        if empty_cell not in tags[row][col]:
            return False
    return True

def find_closest_date_cell(tags, row, col, borders):
    while row > borders.upper_row:
        row -= 1
        if date_cell in tags[row][col]:
            return row, col
    while col > borders.left_col:
        col -= 1
        if date_cell in tags[row][col]:
            return row, col




import pyexcel as pe
import xlrd
import logging


def process_excel(filename):
    if filename.endswith(".csv"):
        # add merged cells here please
        sheet = pe.get_sheet(file_name=filename)
        yield sheet, "name", ()
    else:
        book = pe.get_book(file_name=filename)
        sheets = book.to_dict()
        booklrd = xlrd.open_workbook(filename)
        for name in sheets.keys():
            sheet = book[name]
            fill_merged_cells(sheet, booklrd.sheet_by_name(name).merged_cells)
            yield sheet, name, booklrd.sheet_by_name(name).merged_cells


# filling the empty cells which are located in the merged cells in reality but are read as empty by pyexcel
def fill_merged_cells(sheet, merged_blocks):

    for block in merged_blocks:
        val = sheet[block[0], block[2]]
        for row in range(block[0], block[1]):
            for col in range(block[2], block[3]):
                try:
                    sheet[row, col] = val
                except:
                    # There is a difference in the way xlrd and pyexcel handle merged cells 
                    # Sometimes we end up accessing out of range cells
                    logging.error("Tried to access out of range cell.")
                    break



import unittest
from etk.knowledge_graph.node import Node, URI, BNode, Literal, LiteralType
from etk.etk_exceptions import InvalidGraphNodeValueError, UnknownLiteralType


class TestKGNode(unittest.TestCase):
    def test_nodes_node(self):
        with self.assertRaises(InvalidGraphNodeValueError):
            Node(None)
        # equal
        node1 = Node('1')
        node2 = Node('2')
        self.assertEqual(node1, Node('1'))
        self.assertNotEqual(node1, node2)
        self.assertEqual(hash(node1), hash(Node('1')))
        self.assertNotEqual(hash(node1), hash(node2))
        # value
        self.assertEqual(node1.value, '1')
        # is_valid
        with self.assertRaises(NotImplementedError):
            node1.is_valid()

    def test_nodes_uri(self):
        uri = URI('rdf:type')
        uri2 = URI('rdf:type')
        with self.assertRaises(InvalidGraphNodeValueError):
            URI(None)
        # equal
        self.assertEqual(uri, uri2)
        # hash
        self.assertEqual(len({uri, uri2}), 1)
        # value
        self.assertEqual(uri.value, 'rdf:type')
        # is_valid
        self.assertTrue(uri.is_valid())

    def test_nodes_bnode(self):
        bnode = BNode()
        bnode2 = BNode()
        self.assertNotEqual(bnode, bnode2)
        self.assertEqual(len({bnode, bnode2}), 2)
        self.assertTrue(bnode.is_valid())

        # named bnode
        test_bnode = BNode('test')
        test_bnode2 = BNode('test')
        self.assertEqual(test_bnode, test_bnode2)
        self.assertNotEqual(test_bnode, bnode)
        self.assertEqual(hash(test_bnode), hash(test_bnode2))
        self.assertNotEqual(hash(test_bnode), hash(bnode))
        self.assertTrue(test_bnode.is_valid())

    def test_nodes_literal_type(self):
        xsd_int = LiteralType('xsd:int')
        self.assertEqual(xsd_int, LiteralType.int)
        self.assertEqual(xsd_int.value, LiteralType.int.value)
        complete_xsd_int = LiteralType('http://www.w3.org/2001/XMLSchema#int')
        self.assertEqual(complete_xsd_int, xsd_int)

        rdf_html = LiteralType('rdf:HTML')
        self.assertEqual(rdf_html, LiteralType.HTML)

        for s in (None, '', 'xxx'):
            with self.assertRaises(UnknownLiteralType):
                LiteralType(s)

        with self.assertRaises(UnknownLiteralType):
            LiteralType.xxx

    def test_nodes_literal(self):
        lit = Literal('name', 'en', LiteralType.string)
        lit2 = Literal('nombre', 'es')
        lit3 = Literal('name', 'en', 'xsd:string')
        self.assertNotEqual(lit, lit2)
        self.assertEqual(lit, lit3)
        self.assertEqual(len({lit, lit2, lit3}), 2)
        self.assertEqual(lit.type, LiteralType('xsd:string'))
        self.assertEqual(lit.raw_type, LiteralType.string.value)
        self.assertTrue(lit.is_valid())

    def test_nodes_init_with_nodes(self):
        uri = URI('rdf:type')
        self.assertIsInstance(URI(uri), URI)
        self.assertEqual(URI(uri), uri)

        bnode = BNode()
        self.assertIsInstance(BNode(bnode), BNode)
        self.assertEqual(BNode(bnode), bnode)

        lit = Literal('name', 'en', LiteralType.string)
        self.assertIsInstance(Literal(lit), Literal)
        self.assertEqual(Literal(lit), lit)



from etk.extractors.date_extractor import DateExtractor, DateResolution
from etk.wikidata.value import Precision


def parse_datetime_string(s, *args, **kwargs):
    """
    Automatically convert string to iso datetime string

    :param s:
    :return: iso format datetime string, wikidata date precision
    """

    kwargs['date_value_resolution'] = DateResolution.SECOND

    de = DateExtractor()
    e = de.extract(s, *args, **kwargs)
    if len(e) == 0:
        raise ValueError('No date / datetime detected')

    dt_str = e[0].value
    orig_resolution = de._last_original_resolution

    wd_precision = None
    if orig_resolution == DateResolution.SECOND:
        wd_precision = Precision.second
    if orig_resolution == DateResolution.MINUTE:
        wd_precision = Precision.minute
    if orig_resolution == DateResolution.HOUR:
        wd_precision = Precision.hour
    elif orig_resolution == DateResolution.DAY:
        wd_precision = Precision.day
    elif orig_resolution == DateResolution.MONTH:
        wd_precision = Precision.month
    elif orig_resolution == DateResolution.YEAR:
        wd_precision = Precision.year

        # TODO: support more fine-grind year precisions

    return dt_str, wd_precision
